### PANDAS DATAFRAME MODULE ###

# pypi.org libs
import os
from logging import Logger
from typing import Any, Dict, List, Tuple

import pandas as pd

# local libs
from stpstone._config.global_slots import YAML_MICROSOFT_APPS
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.parsers.folders import DirFilesManagement

# check if this is a windows machine
if os.name == "nt":
    from stpstone.utils.microsoft_apps.excel import DealingExcel


class DealingPd:

    def __init__(self):
        self.dict_sensitivity_labels_office = {
            "public": "e6a9157b-bcf3-4eac-b03e-7cf007ba9fdf",
            "internal": "d5bef5af-bbc1-4d24-bb43-47a55a90f763",
            "confidential": "f33522dc-133a-44f0-ba7f-cdd6f493ffbd",
            "restricted": "69878aea-41a3-47a7-b03f-1211df6e7785"
        }

    def append_df_to_Excel(
        self,
        filename: str,
        list_tup_df_sheet_name: List[Tuple[pd.DataFrame, str]],
        bl_header: bool = True,
        bl_index: int = 0,
        mode: str = "w",
        label_sensitivity: str = "internal",
        bl_set_sensitivity_label: bool = False,
        bl_debug_mode: bool = False,
    ) -> None:
        """
        Append dataframe to Excel

        Args:
            filename (str): Path to Excel file
            list_tup_df_sheet_name (List[Tuple[pd.DataFrame, str]]): List of tuples with dataframe and sheet name
            bl_header (bool): Whether to write headers
            bl_index (int): Whether to write index
            mode (str): Mode to open Excel file
            label_sensitivity (str): Label sensitivity
            bl_set_sensitivity_label (bool): Whether to set sensitivity label
            bl_debug_mode (bool): Whether to print debug messages

        Returns:
            None
        """
        if bl_debug_mode == True:
            print("LIST_TUP_DF_SHEET_NAME: {}".format(list_tup_df_sheet_name))
            print("FILENAME: {}".format(filename))
            print("MODE: {}".format(mode))
            print("BL_INDEX: {}".format(bl_index))
        with pd.ExcelWriter(filename, engine="xlsxwriter", mode=mode) as writer:
            for df_, sheet_name in list_tup_df_sheet_name:
                if bl_index == 0:
                    df_.reset_index(drop=True, inplace=True)
                df_.to_excel(writer, sheet_name, index=bl_index, header=bl_header)
        if (bl_set_sensitivity_label == True) and (os.name == "nt"):
            DealingExcel().xlsx_sensitivity_label(
                filename,
                self.dict_sensitivity_labels_office,
                label_sensitivity.capitalize(),
            )

    def export_xl(
        self,
        logger: Logger,
        path_xlsx: str,
        list_tup_df_sheet_name: List[Tuple[pd.DataFrame, str]],
        range_columns: str = "A:CC",
        bl_adjust_layout: bool = False,
        bl_debug_mode: bool = False,
    ) -> bool:
        """
        Export dataframe to Excel

        Args:
            - logger (logging.Logger): Logger object
            - path_xlsx (str): Path to Excel file
            - list_tup_df_sheet_name (list): List of tuples of DataFrames and sheet names
            - range_columns (str): Range of columns to autofit
            - bl_adjust_layout (bool): Whether to adjust layout of Excel file
            - bl_debug_mode (bool): Whether to print debug messages

        Returns:
            bool
        """
        self.append_df_to_Excel(
            path_xlsx,
            list_tup_df_sheet_name,
            bl_header=True,
            bl_index=0,
            bl_debug_mode=bl_debug_mode,
        )
        blame_xpt = DirFilesManagement().object_exists(path_xlsx)
        if blame_xpt == True:
            if bl_adjust_layout == True:
                for _, sheet_name in list_tup_df_sheet_name:
                    xl_app, wb = DealingExcel().open_xl(path_xlsx)
                    DealingExcel().autofit_range_columns(
                        sheet_name, range_columns, xl_app, wb
                    )
                    DealingExcel().close_wb(wb)
        else:
            CreateLog().warning(
                logger, "File not saved to hard drive: {}".format(path_xlsx)
            )
            raise Exception("File not saved to hard drive: {}".format(path_xlsx))
        return blame_xpt

    def settingup_pandas(
        self,
        int_decimal_places: int = 3,
        bl_wrap_repr: bool = False,
        int_max_rows: int = 25,
    ) -> None:
        """
        Setting up pandas options

        Args:
            int_decimal_places (int): Number of decimal places to display in output
            bl_wrap_repr (bool): Whether to wrap repr(DataFrame) across additional lines
            int_max_rows (int): Maximum number of rows to display in output

        Returns:
            None
        """
        pd.set_option("display.precision", int_decimal_places)
        pd.set_option("display.expand_frame_repr", bl_wrap_repr)
        pd.set_option("display.max_rows", int_max_rows)

    def convert_datetime_columns(
        self,
        df_: pd.DataFrame,
        list_col_date: List[str],
        bl_pandas_convertion: bool = True,
    ) -> pd.DataFrame:
        """
        Convert datetime columns

        Args:
            df_ (pandas.DataFrame): DataFrame to convert
            list_col_date (list): List of columns to convert
            bl_pandas_convertion (bool): Whether to use pandas conversion or excel format

        Returns:
            pd.DataFrame
        """
        # checking wheter to covert through a pandas convertion, or resort to a excel format
        #   transformation of data in date column format
        if bl_pandas_convertion:
            for col_date in list_col_date:
                df_.loc[:, col_date] = pd.to_datetime(df_[col_date], unit="s").dt.date
        else:
            # corventing list column dates to string type
            df_ = df_.astype(dict(zip(list_col_date, [str] * len(list_col_date))))
            # looping through each row
            for index, row in df_.iterrows():
                for col_date in list_col_date:
                    if "-" in row[col_date]:
                        ano = int(row[col_date].split(" ")[0].split("-")[0])
                        mes = int(row[col_date].split(" ")[0].split("-")[1])
                        dia = int(row[col_date].split(" ")[0].split("-")[2])
                        df_.loc[index, col_date] = DatesBR().build_date(ano, mes, dia)
                    else:
                        df_.loc[index, col_date] = DatesBR().excel_float_to_date(
                            int(row[col_date])
                        )
        # returning dataframe with date column to datetime forma
        return df_

    def merge_dfs_into_df(
        self, df_1: pd.DataFrame, df_2: pd.DataFrame, list_cols: List[str]
    ) -> pd.DataFrame:
        """
        Merging two dataframes and removing their intersections into a list of columns

        Args:
            df_1 (pandas.DataFrame): First dataframe to merge
            df_2 (pandas.DataFrame): Second dataframe to merge
            list_cols (list): List of columns to merge

        Returns:
            pd.DataFrame
        """
        df_intersec = pd.merge(df_1, df_2, how="inner", on=list_cols)
        df_merge = pd.merge(
            df_2, df_intersec, how="outer", on=list_cols, indicator=True
        )
        df_merge = df_merge.loc[df_merge._merge == "left_only"]
        df_merge.dropna(how="all", axis=1, inplace=True)
        try:
            df_merge = df_merge.drop(columns="_merge")
        except:
            pass
        for column in df_merge.columns:
            if column in list_cols:
                pass
            else:
                df_merge = df_merge.rename(columns={str(column): str(column)[:-2]})
        return df_merge

    def max_chrs_per_column(df_: pd.DataFrame) -> Dict[str, int]:
        """
        Calculate the maximum number of characters per column

        Args:
            df_ (pandas.DataFrame): DataFrame to calculate

        Returns:
            Dict[str, int
        """
        dict_ = dict()
        for col_ in list(df_.columns):
            dict_[col_] = df_[col_].astype(str).str.len().max()
        return dict_



import logging
import json
from typing import Optional, Union, List, Dict
from webdriver_manager.chrome import ChromeDriverManager
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.remote.webelement import WebElement


class SeleniumWD:

    def __init__(
        self,
        url: str,
        path_webdriver: Optional[str] = None,
        int_port: Optional[int] = None,
        str_user_agent: Optional[str] = None,
        int_wait_load_seconds: int = 10,
        int_delay_seconds: int = 10,
        str_proxy: Optional[str] = None,
        bl_headless: bool = False,
        bl_incognito: bool = False,
        list_args: Optional[List[str]] = None
    ) -> None:
        """
        Initialization of selenium web driver

        Args:
            url (str): url to open
            path_webdriver (str, optional): path to webdriver. Defaults to None.
            int_port (int, optional): port to open. Defaults to None.
            str_user_agent (str, optional): user agent. Defaults to "Mozilla/5.0 (Windowns NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36".
            int_wait_load_seconds (int, optional): time to wait for page to load. Defaults to 10.
            int_delay_seconds (int, optional): time to wait between actions. Defaults to 10.
            str_proxy (str, optional): proxy to use. Defaults to None.
            bl_opn_min (bool, optional): open in minimal mode. Defaults to False.
            bl_headless (bool, optional): open in headless mode. Defaults to False.
            bl_incognito (bool, optional): open in incognito mode. Defaults to False.
            list_args (Optional[List[str]], optional): webdriver arguments. Defaults to None.

        Returns:
            None

        Notes:
            User agents: https://gist.github.com/pzb/b4b6f57144aea7827ae4
            Webdriver arguments: https://chromedriver.chromium.org/capabilities
            Chromium command line switches: https://gist.github.com/dodying/34ea4760a699b47825a766051f47d43b
        """
        self.url = url
        self.path_webdriver = path_webdriver
        self.int_port = int_port
        self.str_user_agent = str_user_agent if str_user_agent is not None \
            else "Mozilla/5.0 (Windowns NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36"
        self.int_wait_load_seconds = int_wait_load_seconds
        self.int_delay_seconds = int_delay_seconds
        self.bl_headless = bl_headless
        self.bl_incognito = bl_incognito
        self.list_default_args = list_args if list_args is not None else [
            "--no-sandbox",
            "--disable-gpu",
            "--disable-setuid-sandbox",
            "--disable-web-security",
            "--disable-dev-shm-usage",
            "--memory-pressure-off",
            "--ignore-certificate-errors",
            "--disable-features=site-per-process",
            "--disable-extensions",
            "--disable-popup-blocking",
            "--disable-notifications",
            "--window-size=1920,1080",
            "--window-position=0,0",
            "--enable-unsafe-swiftshader",
            f"--user-agent={str_user_agent}"
        ]
        # set headless mode for operations without graphical user interface (GUI) - if true
        if self.bl_headless == True:
            self.list_default_args.append("--headless=new")
        if self.bl_incognito == True:
            self.list_default_args.append("--incognito")
        if str_proxy is not None:
            self.list_default_args.append(f"--proxy-server={str_proxy}")
        self.web_driver = self.get_web_driver

    @property
    def get_web_driver(self) -> WebDriver:
        d = DesiredCapabilities.CHROME
        d["goog:loggingPrefs"] = {"performance": "ALL"}
        browser_options = webdriver.ChromeOptions()
        for arg in self.list_default_args:
            browser_options.add_argument(arg)
        if (self.path_webdriver is None) and (self.int_port is not None):
            service = Service(executable_path=ChromeDriverManager().install())
        else:
            service = Service(executable_path=self.path_webdriver)
        web_driver = webdriver.Chrome(service=service, options=browser_options)
        web_driver.get(self.url)
        web_driver.implicitly_wait(self.int_wait_load_seconds)
        return web_driver

    def process_log(self, log:Dict[str, Union[str, dict]]) -> Optional[Dict[str, Union[str, dict]]]:
        log = json.loads(log["message"])["message"]
        if ("Network.response" in log["method"] and "params" in log.keys()):
            body = self.web_driver.execute_cdp_cmd("Network.getResponseBody", {"requestId": log[
                "params"]["requestId"]})
            print(json.dumps(body, indent=4, sort_keys=True))
            return log["params"]

    @property
    def get_browser_log_entries(self) -> List[Dict[str, Union[str, dict]]]:
        loglevels = {"NOTSET": 0, "DEBUG": 10, "INFO": 20,
                     "WARNING": 30, "ERROR": 40, "SEVERE": 40, "CRITICAL": 50}
        browserlog = logging.getLogger("chrome")
        slurped_logs = self.web_driver.get_log("web_driver")
        for entry in slurped_logs:
            # convert broswer log to python log format
            rec = browserlog.makeRecord("%s.%s" % (browserlog.name, entry["source"]), loglevels.get(
                entry["level"]), ".", 0, entry["message"], None, None)
            # log using original timestamp.. us -> ms
            rec.created = entry["timestamp"] / 1000
            try:
                # add web_driver log to python log
                browserlog.handle(rec)
            except:
                print(entry)
        return slurped_logs

    def process_browser_log_entry(self, entry: Dict[str, Union[str, dict]]) \
        -> Dict[str, Union[str, dict]]:
        return json.loads(entry["message"])["message"]

    @property
    def get_network_traffic(self) -> List[Dict[str, Union[str, dict]]]:
        browser_log = self.web_driver.get_log("performance")
        list_events = [self.process_browser_log_entry(entry) for entry in browser_log]
        list_events = [event for event in list_events if "Network.response" in event["method"]]
        return list_events

    def find_element(self, selector:Union[WebElement, WebDriver], str_element_interest:str,
                     selector_type:str="XPATH") -> WebElement:
        return selector.find_element(getattr(By, selector_type), str_element_interest)

    def find_elements(self, selector:Union[WebElement, WebDriver], str_element_interest:str,
                     selector_type:str="XPATH") -> WebElement:
        return selector.find_elements(getattr(By, selector_type), str_element_interest)

    def fill_input(self, web_element:WebElement, str_input:str) -> None:
        web_element.send_keys(str_input)

    def el_is_enabled(self, str_xpath:str) -> bool:
        return ec.presence_of_element_located((By.XPATH, str_xpath))

    def wait_until_el_loaded(self, str_xpath:str) -> WebDriverWait:
        return WebDriverWait(self.web_driver, self.int_delay_seconds).until(self.el_is_enabled(str_xpath))

    def wait(self, seconds:int) -> None:
        self.web_driver.implicitly_wait(seconds)



import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Tuple, Dict, Union
from numbers import Number
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from lxml.html import HtmlElement
from stpstone._config.global_slots import YAML_BMF_INTEREST_RATES
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.loggs.create_logs import CreateLog


class BMFInterestRates(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_BMF_INTEREST_RATES,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs
        self.dt_ref_ddmmyyyy = dt_ref.strftime('%d/%m/%Y')
        self.dt_ref_yyyymmdd = dt_ref.strftime('%Y%m%d')

    def td_th_parser(self, root: HtmlElement, int_iter: int) -> List[Dict[str, Union[str, Number]]]:
        list_th = list()
        for x in HtmlHandler().lxml_xpath(
            root, YAML_BMF_INTEREST_RATES['rates']['xpaths']['list_th'].format(int_iter)
        ):
            str_raw = StrHandler().remove_diacritics(x.text.strip().lower())
            if str_raw == "dias":
                list_th.append("DIAS_CORRIDOS")
            elif str_raw == "di x pre":
                list_th.append("DI_PRE_252")
                list_th.append("DI_PRE_360")
            elif str_raw == "selic x pre":
                list_th.append("SELIC_PRE_252")
            elif str_raw == "di x tr":
                list_th.append("DI_TR_252")
                list_th.append("DI_TR_360")
            elif str_raw == "dolar x pre":
                list_th.append("DOLAR_PRE_252")
                list_th.append("DOLAR_PRE_360")
            elif str_raw == "real x euro":
                list_th.append("REAL_EURO")
            elif str_raw == "di x euro":
                list_th.append("DI_EURO_360")
            elif str_raw == "tbf x pre":
                list_th.append("TBF_PRE_252")
                list_th.append("TBF_PRE_360")
            elif str_raw == "tr x pre":
                list_th.append("TR_PRE_252")
                list_th.append("TR_PRE_360")
            elif str_raw == "di x dolar":
                list_th.append("DI_DOLAR_360")
            elif str_raw == "cupom cambial oc1":
                list_th.append("CUPOM_CAMBIAL_OC1_360")
            elif str_raw == "cupom limpo":
                list_th.append("CUPOM_LIMPO_360")
            elif str_raw == "real x dolar":
                list_th.append("REAL_DOLAR")
            elif str_raw == "ibrx-50":
                list_th.append("IBRX_50")
            elif str_raw == "ibovespa":
                list_th.append("IBOVESPA")
            elif str_raw == "di x igp-m":
                list_th.append("DI_IGP_M_252")
            elif str_raw == "di x ipca":
                list_th.append("DI_IPCA_252")
            elif str_raw == "ajuste pre":
                list_th.append("AJUSTE_PRE_252")
                list_th.append("AJUSTE_PRE_360")
            elif str_raw == "ajuste cupom":
                list_th.append("AJUSTE_CUPOM_360")
            elif str_raw == "real x iene":
                list_th.append("REAL_IENE")
            elif str_raw == "spread libor euro x dolar":
                list_th.append("SPREAD_LIBOR_EURO_DOLAR")
            elif str_raw == "libor":
                list_th.append("LIBOR_360")
            else:
                raise Exception(f"Table Header not Found: {str_raw}")
        list_td = [
            float(x.text.strip().replace(",", "."))
            for x in HtmlHandler().lxml_xpath(
                root, YAML_BMF_INTEREST_RATES['rates']['xpaths']['list_td'].format(int_iter)
            )
        ]
        list_ser = HandlingDicts().pair_headers_with_data(list_th, list_td)
        return list_ser

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        list_dfs = list()
        bl_debug = True if StrHandler().match_string_like(
            resp_req.url, '*bl_debug=True*') == True else False
        root = HtmlHandler().lxml_parser(resp_req)
        # export html tree to data folder, if is user's will
        if bl_debug == True:
            path_project = DirFilesManagement().find_project_root(marker='pyproject.toml')
            HtmlHandler().html_tree(root, file_path=rf'{path_project}/data/test.html')
        for i in range(1, 5):
            list_ser = self.td_th_parser(root, i)
            list_dfs.append(pd.DataFrame(list_ser))
        df_ = list_dfs[0]
        for i in range(1, len(list_dfs)):
            df_ = df_.merge(list_dfs[i], how='left', on='DIAS_CORRIDOS', suffixes=('', f'_{i}'))
        return df_



from datetime import date
from typing import Optional
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class DebenturesBR(metaclass=TypeChecker):
    """
    A comprehensive class for pricing and analyzing Brazilian debentures,
    excluding float_duration-related calculations which are already implemented.
    """
    
    def __init__(
        self,
        float_fv: float = 1000,
        float_coupon_r: float = 0.1,
        int_coupon_freq: int = 2,
        int_maturity_years: int = 5,
        float_yield: float = 0.12,
        float_tax_rate: float = 0.15
    ) -> None:
        """
        Initialize the debenture with default or provided parameters.
        
        Args:
            float_fv: Nominal value (valor de face)
            float_coupon_r: Annual coupon rate (decimal)
            int_coupon_freq: Payments per year
            int_maturity_years: Years to maturity
            float_yield: Market yield (decimal)
            float_tax_rate: Tax on interest (decimal)
        """
        self.float_fv = float_fv
        self.float_coupon_r = float_coupon_r
        self.int_coupon_freq = int_coupon_freq
        self.int_maturity_years = int_maturity_years
        self.float_yield = float_yield
        self.float_tax_rate = float_tax_rate
    
    def calculate_price(self) -> float:
        """
        Calculate the theoretical float_price of the debenture considering tax on coupons.
        
        Returns:
            Present value/float_price of the debenture
        """
        float_price: float = 0.0
        int_periods: int = int(self.int_maturity_years * self.int_coupon_freq)
        float_period_rate: float = self.float_yield / self.int_coupon_freq
        float_float_coupon_payment: float = (self.float_fv * self.float_coupon_r / 
                               self.int_coupon_freq) * (1 - self.float_tax_rate)
        for t in range(1, int_periods + 1):
            float_cf: float = float_float_coupon_payment + self.float_fv if t == int_periods \
                else float_float_coupon_payment
            float_price += float_cf / ((1 + float_period_rate) ** t)
        return float_price
    
    def current_yield(self, float_market_price: Optional[float] = None) -> float:
        """
        Calculate current yield (annual coupon payment / current market float_price).
        
        Args:
            float_market_price: Current market float_price (if None, uses calculated float_price)
            
        Returns:
            Current yield (decimal)
        """
        float_price: float = float_market_price if float_market_price is not None \
            else self.calculate_price()
        float_annual_coupon: float = self.float_fv * self.float_coupon_r * (1 - self.float_tax_rate)
        return float_annual_coupon / float_price
    
    def capital_gain_yield(self, float_market_price: Optional[float] = None) -> float:
        """
        Calculate expected capital gain/loss yield to maturity.
        
        Args:
            float_market_price: Current market float_price (if None, uses calculated float_price)
            
        Returns:
            Capital gain yield (decimal)
        """
        float_price: float = float_market_price if float_market_price is not None \
            else self.calculate_price()
        return (self.float_fv - float_price) / float_price
    
    def float_total_return(self, float_market_price: Optional[float] = None) -> float:
        """
        Calculate total expected return to maturity (current yield + capital gain).
        
        Args:
            float_market_price: Current market float_price (if None, uses calculated float_price)
            
        Returns:
            Total expected return (decimal)
        """
        return (self.current_yield(float_market_price) + 
                self.capital_gain_yield(float_market_price))

    def accrued_interest(
        self,
        dt_settlement: Optional[date] = None,
        dt_last_coupon: Optional[date] = None,
        dt_next_coupon: Optional[date] = None,
        str_day_count_convention: str = "actual/actual"
    ) -> float:
        """
        Calculate float_accrued interest since last coupon payment.
        
        Args:
            dt_settlement: Date for which to calculate float_accrued interest (default: today)
            dt_last_coupon: Date of last coupon payment (if not provided, calculated)
            dt_next_coupon: Date of next coupon payment (if not provided, calculated)
            str_day_count_convention: Day count method ("actual/actual", "30/360", etc.)
            
        Returns:
            float: Accrued interest amount
            
        Raises:
            ValueError: If dates are invalid or day count convention not supported
        """
        if dt_settlement is None:
            dt_settlement = date.today()
        if dt_last_coupon is None or dt_next_coupon is None:
            last_cp, next_cp = self._calculate_coupon_dates(dt_settlement)
            dt_last_coupon = dt_last_coupon or last_cp
            dt_next_coupon = dt_next_coupon or next_cp
        if not (dt_last_coupon <= dt_settlement <= dt_next_coupon):
            raise ValueError("Settlement date must be between coupon dates")
        # calculate days according to convention
        if str_day_count_convention.lower() == "actual/actual":
            int_days_accrued = (dt_settlement - dt_last_coupon).days
            int_days_in_period = (dt_next_coupon - dt_last_coupon).days
        elif str_day_count_convention.lower() == "30/360":
            int_days_accrued = self._days_360(dt_last_coupon, dt_settlement)
            int_days_in_period = self._days_360(dt_last_coupon, dt_next_coupon)
        else:
            raise ValueError(f"Unsupported day count convention: {str_day_count_convention}")
        float_coupon_payment = (self.float_fv * self.float_coupon_r) / self.int_coupon_freq
        float_accrued = float_coupon_payment * (int_days_accrued / int_days_in_period)
        # apply tax if this is a taxable coupon
        if self.float_tax_rate > 0:
            float_accrued *= (1 - self.float_tax_rate)
        return float_accrued
    
    def _calculate_coupon_dates(self, dt_settlement: date) -> tuple[date, date]:
        """
        Calculate previous and next coupon dates based on settlement date.
        
        Args:
            dt_settlement: Reference date for calculation
            
        Returns:
            tuple: (dt_last_coupon, dt_next_coupon)
        """
        # this is a simplified implementation - would need actual issue date in real usage
        int_months_between = 12 // self.int_coupon_freq
        # find most recent coupon date before settlement
        dt_last_coupon = date(dt_settlement.year, 1, 1)
        while dt_last_coupon <= dt_settlement:
            dt_last_coupon = dt_last_coupon.replace(month=dt_last_coupon.month + int_months_between)
        dt_last_coupon = dt_last_coupon.replace(month=dt_last_coupon.month - int_months_between)
        # next coupon is int_months_between after last coupon
        dt_next_coupon = dt_last_coupon.replace(month=dt_last_coupon.month + int_months_between)
        return dt_last_coupon, dt_next_coupon
    
    def _days_360(self, dt_start: date, dt_end: date) -> int:
        """
        Calculate days between dates using 30/360 convention.
        
        Args:
            dt_start: Start date
            dt_end: End date
            
        Returns:
            int: Days between dates under 30/360 convention
        """
        dt_d1 = min(dt_start.day, 30)
        dt_d2 = min(dt_end.day, 30) if dt_d1 == 30 else dt_end.day
        return (dt_end.year - dt_start.year) * 360 + \
               (dt_end.month - dt_start.month) * 30 + \
               (dt_d2 - dt_d1)

    def dirty_price(
        self, 
        float_clean_price: Optional[float] = None,
        dt_settlement: Optional[date] = None,
        dt_last_coupon: Optional[date] = None,
        dt_next_coupon: Optional[date] = None,
        str_day_count_convention: str = "actual/actual"
    ) -> float:
        """
        Calculate dirty price (clean price + accrued interest).
        
        Args:
            float_clean_price: Clean price (if None, uses calculated price)
            dt_settlement: Settlement date for accrued interest calculation
            dt_last_coupon: Last coupon payment date
            dt_next_coupon: Next coupon payment date
            str_day_count_convention: Day count convention method
            
        Returns:
            Dirty price including accrued interest
        """
        clean_price: float = float_clean_price if float_clean_price is not None \
            else self.calculate_price()
        accrued = self.accrued_interest(
            dt_settlement=dt_settlement,
            dt_last_coupon=dt_last_coupon,
            dt_next_coupon=dt_next_coupon,
            str_day_count_convention=str_day_count_convention
        )
        return clean_price + accrued
    
    def holding_period_return(
        self,
        float_purchase_price: float,
        float_sale_price: float,
        int_holding_period: float
    ) -> float:
        """
        Calculate annualized holding period return.
        
        Args:
            float_purchase_price: Price when bought
            float_sale_price: Price when sold
            int_holding_period: Holding period in years
            
        Returns:
            Annualized return (decimal)
        """
        float_total_return: float = (float_sale_price - float_purchase_price) / float_purchase_price
        return (1 + float_total_return) ** (1 / int_holding_period) - 1
    
    def credit_spread(self, float_risk_free_rate: float) -> float:
        """
        Calculate credit spread over risk-free rate.
        
        Args:
            float_risk_free_rate: Risk-free rate (decimal)
            
        Returns:
            Credit spread (decimal)
        """
        return self.float_yield - float_risk_free_rate
    
    def early_redemption_value(self, float_call_price: float, float_call_date: float) -> float:
        """
        Calculate the minimum between the debenture's present value to maturity and
        its present value if called at the call date, considering all cash flows.
        
        Args:
            float_call_price: Price at which the debenture can be called (% of face value)
            float_call_date: Time until call date in years (must be <= maturity)
            
        Returns:
            float: Minimum value between holding to maturity and being called
            
        Raises:
            ValueError: If float_call_date is after maturity or negative
        """
        float_call_value = 0.0
        if float_call_date > self.int_maturity_years:
            raise ValueError("Call date cannot be after maturity")
        if float_call_date <= 0:
            raise ValueError("Call date must be positive")
        # value if held to maturity
        float_maturity_value = self.calculate_price()
        # present value if called
        int_call_periods = int(float_call_date * self.int_coupon_freq)
        float_period_rate = self.float_yield / self.int_coupon_freq
        float_coupon_payment = (self.float_fv * self.float_coupon_r / 
                        self.int_coupon_freq) * (1 - self.float_tax_rate)
        for t in range(1, int_call_periods + 1):
            if t == int_call_periods:
                # at call date, investor receives call float_price plus final coupon
                float_cf = float_coupon_payment + float_call_price
            else:
                float_cf = float_coupon_payment
            
            float_call_value += float_cf / ((1 + float_period_rate) ** t)
        
        return min(float_maturity_value, float_call_value)
    
    def inflation_adjusted_return(self, float_inflation_rate: float) -> float:
        """
        Calculate real return adjusted for inflation.
        
        Args:
            float_inflation_rate: Expected inflation rate (decimal)
            
        Returns:
            Real return (decimal)
        """
        nominal_return: float = self.float_yield
        return (1 + nominal_return) / (1 + float_inflation_rate) - 1
    
    def yield_to_call(
        self,
        float_call_price: float,
        float_call_date: float,
        float_current_price: float,
        int_max_iterations: int = 100,
        float_precision: float = 1e-6
    ) -> float:
        """
        Calculate yield to call using Newton-Raphson method.
        
        Args:
            float_call_price: Price at which debenture can be called (% of face value)
            float_call_date: Time until call date in years
            float_current_price: Current market float_price of the debenture
            int_max_iterations: Maximum iterations for convergence
            float_precision: Desired float_precision level
            
        Returns:
            float: Yield to call (annualized)
            
        Raises:
            ValueError: If float_call_date is invalid or calculation doesn't converge
        """
        if float_call_date <= 0:
            raise ValueError("Call date must be positive")
        if float_call_date > self.int_maturity_y:
            raise ValueError("Call date cannot be after maturity")
        if float_current_price <= 0:
            raise ValueError("Current float_price must be positive")
        float_ytc = self.float_coupon_r
        int_periods = int(float_call_date * self.int_coupon_freq)
        float_coupon_payment = (self.float_fv * self.float_coupon_r / 
                         self.int_coupon_freq) * (1 - self.float_tax_rate)
        for _ in range(int_max_iterations):
            # calculate float_price and derivative at current guess
            float_price = 0.0
            float_duration = 0.0
            for t in range(1, int_periods + 1):
                if t == int_periods:
                    float_cash_flow = float_coupon_payment + float_call_price
                else:
                    float_cash_flow = float_coupon_payment
                float_discount_rate = (1 + float_ytc/self.int_coupon_freq) ** t
                float_price += float_cash_flow / float_discount_rate
                float_duration += t * float_cash_flow / (float_discount_rate * \
                                                   (1 + float_ytc/self.int_coupon_freq))
            # newton-Raphson update
            price_diff = float_price - float_current_price
            if abs(price_diff) < float_precision:
                return float_ytc
            # modified float_duration
            float_mod_duration = float_duration / float_price
            # update guess
            float_ytc -= price_diff / (-float_mod_duration * float_price)
            # ensure yield stays reasonable - prevent negative yields < -99%
            float_ytc = max(float_ytc, -0.99)
        raise ValueError("Yield to call calculation did not converge after "\
                         + f"{int_max_iterations} iterations")

    def _calculate_ytc_price(
        self,
        ytc_rate: float,
        float_call_price: float,
        call_periods: int
    ) -> float:
        """
        Helper method to calculate float_price given a yield-to-call rate.
        
        Args:
            ytc_rate: Yield to call rate to test
            float_call_price: Call float_price
            call_periods: Number of int_periods until call
            
        Returns:
            float: Calculated float_price
        """
        float_price = 0.0
        float_coupon_payment = (self.float_fv * self.float_coupon_r / 
                         self.int_coupon_freq) * (1 - self.float_tax_rate)
        for t in range(1, call_periods + 1):
            if t == call_periods:
                float_cash_flow = float_coupon_payment + float_call_price
            else:
                float_cash_flow = float_coupon_payment
            float_price += float_cash_flow / ((1 + ytc_rate/self.int_coupon_freq) ** t)
        return float_price
    
    def tax_adjusted_yield(self, float_investor_tax_rate: float) -> float:
        """
        Adjust yield for investor's specific tax rate.
        
        Args:
            float_investor_tax_rate: Investor's marginal tax rate (decimal)
            
        Returns:
            After-tax yield (decimal)
        """
        return self.float_yield * (1 - float_investor_tax_rate)


import pycountry
import pytz
import pycountry_convert as pc
from datetime import datetime
from typing import List, Optional, Dict, Literal
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class WWTimezones(metaclass=TypeChecker):

    def get_timezones_by_country_code(self, country_code: str) -> Optional[List[str]]:
        country = pycountry.countries.get(alpha_2=country_code.upper())
        return pytz.country_timezones.get(country.alpha_2) if country else None

    def get_countries_in_timezone(self, timezone_name: str) -> List[str]:
        return [
            code for code, tz_list in pytz.country_timezones.items()
            if timezone_name in tz_list
        ]

    def get_current_time_in_country(self, country_code: str, int_tz: int = 0) -> Optional[str]:
        timezones = self.get_timezones_by_country_code(country_code.upper())
        if not timezones:
            return None
        tz = pytz.timezone(timezones[int_tz])
        return datetime.now(tz).isoformat(sep=" ")

    def get_country_from_timezone(self, timezone_name: str) -> List[str]:
        return self.get_countries_in_timezone(timezone_name)

    @property
    def get_all_timezones_grouped(self) -> Dict[str, List[str]]:
        return pytz.country_timezones


class WWGeography(metaclass=TypeChecker):

    def get_country_details(self, country_code: str) -> Optional[Dict[str, str]]:
        country = (
            pycountry.countries.get(alpha_2=country_code.upper())
            or pycountry.countries.get(alpha_3=country_code.upper())
        )
        if not country:
            return None
        return {
            "name": country.name,
            "alpha_2": country.alpha_2,
            "alpha_3": country.alpha_3,
            "official_name": getattr(country, "official_name", ""),
        }

    def bl_valid_country_code(self, country_code: str) -> bool:
        return bool(
            pycountry.countries.get(alpha_2=country_code.upper()) or
            pycountry.countries.get(alpha_3=country_code.upper())
        )

    def get_country_flag_emoji(self, country_code: str) -> Optional[str]:
        if self.bl_valid_country_code(country_code.upper()) == False:
            return None
        return "".join(chr(ord(c) + 127397) for c in country_code.upper())

    def get_country_details_by_name(self, name: str) -> Optional[Dict[str, str]]:
        country = pycountry.countries.search_fuzzy(name)
        if country:
            return self.get_country_details(country[0].alpha_2)
        return None

    def get_continent_by_country_code(self, country_code: str) -> Optional[str]:
        if self.bl_valid_country_code(country_code) == False:
            return None
        try:
            continent_code = pc.country_alpha2_to_continent_code(country_code.upper())
            return pc.convert_continent_code_to_continent_name(continent_code)
        except (KeyError, ValueError):
            return None

    def get_continent_code_by_country_code(self, country_code: str) \
        -> Literal["AF", "AS", "EU", "NA", "SA", "OC", "AN"]:
        if not self.bl_valid_country_code(country_code):
            return None
        try:
            return pc.country_alpha2_to_continent_code(country_code.upper()).upper()
        except (KeyError, ValueError):
            return None



### HANDLING ONEDRIVE ISSUES ###

from getpass import getuser
from stpstone.utils.parsers.folders import DirFilesManagement


class OneDrive:

    @property
    def check_sync_status(self,
                          dir_path_business='C:\\Users\\{}\\AppData\\Local\\Microsoft\\OneDrive\\logs\\Business1\\',
                          name_like='SyncEngine-*', bl_to_datetime=True):
        """
        DOCSTRING: CHECK WHETHER SYNC SERVICE IN LOCAL MACHINE IS ALIVE
        INPUTS: -
        OUTPUTS: DATETIME
        """
        # definifing local dir_path_business to status file
        dir_path_business = dir_path_business.format(getuser())
        # catch the most recent edited file
        complete_status_file_path = DirFilesManagement().choose_last_saved_file_w_rule(
            dir_path_business, name_like)
        # return last edition datetime
        return DirFilesManagement().time_last_edition(complete_status_file_path,
                                                      bl_to_datetime=bl_to_datetime)



### HANDLING DICTIONARIES ISSUES ###

import re
from collections import Counter, OrderedDict, defaultdict
from functools import cmp_to_key
from heapq import nlargest, nsmallest
from itertools import groupby
from numbers import Number
from operator import itemgetter
from typing import Any, Callable, Dict, List, Optional, Union


class HandlingDicts:

    def min_val_key(self, dict_active):
        """
        DOCSTRING: MINIMUN VALUE FOR A GIVEN SET OF VALUES IN A DICTIONARY
        INPUTS: ACTIVE DICTIONARY
        OUTPUTS: KEY, VALUE
        """
        return min(dict_active.items(), key=itemgetter(1))

    def max_val_key(self, dict_active):
        """
        DOCSTRING: MAXIMUN VALUE FOR A GIVEN SET OF VALUES IN A DICTIONARY
        INPUTS: ACTIVE DICTIONARY
        OUTPUTS: KEY, VALUE
        """
        return max(dict_active.items(), key=itemgetter(1))

    def merge_n_dicts(self, *dicts):
        """
        DOCSTRING: MERGE DICTIONARIES, FOR PYTHON 3.5+
        INPUTS: DICTIONARIES
        OUTPUTS: DICTIONARY
        """
        dict_xpt = dict()
        for dict_ in dicts:
            dict_xpt = {**dict_xpt, **dict_}
        return dict_xpt

    def cmp(self, x, y):
        """
        Replacement for built-in function cmp that was removed in Python 3
        Compare the two objects x and y and return an integer according to
        the outcome. The return value is negative if x < y, zero if x == y
        and strictly positive if x > y.
        https://portingguide.readthedocs.io/en/latest/comparisons.html#the-cmp-function
        """
        return (x > y) - (x < y)

    def multikeysort(self, items, columns):
        """
        REFERENCES: https://stackoverflow.com/questions/1143671/how-to-sort-objects-by-multiple-keys-in-python,
            https://stackoverflow.com/questions/28502774/typeerror-cmp-is-an-invalid-keyword-argument-for-this-function
        DOCSTRING: SORT A LIST OF DICTIONARIES
        INPUTS: LIST OF DICTS AND LIST OF COLUMNS, IF THERE IS A NEGATIVE (-) SIGN ON KEY,
            IT WIL BE ORDERED IN REVERSE
        OUTPUTS: LIST OF DICTIONARIES
        """
        comparers = [
            (
                (itemgetter(col[1:].strip()), -1)
                if col.startswith("-")
                else (itemgetter(col.strip()), 1)
            )
            for col in columns
        ]
        def comparer(left, right):
            comparer_iter = (
                self.cmp(fn(left), fn(right)) * mult for fn, mult in comparers
            )
            return next((result for result in comparer_iter if result), 0)
        return sorted(items, key=cmp_to_key(comparer))

    def sum_values_selected_keys(self, list_ser, list_keys_merge=None, bl_sum_values_key=True):
        """
        DOCSTRING: MERGE DICTS FOR EVERY KEY REPETITION
        INPUTS: FOREIGNER KEY, DICTS
        OUTPUTS: DICTIONARY
        """
        # setting default variables
        dict_export = defaultdict(list)
        list_counter_dicts = list()
        # if list of keys to merge is none, return a list of every values for the same key
        if list_keys_merge != None:
            # iterating through dictionaries of interest an merging accordingly to foreigner key
            for dict_ in list_ser:
                for key, value in dict_.items():
                    if key in list_keys_merge:
                        dict_export[key].append(value)
                    else:
                        dict_export[key] = value
            if bl_sum_values_key == True:
                return {
                    k: (sum(v) if isinstance(v, list) else v)
                    for k, v in dict_export.items()
                }
            else:
                return dict_export
        else:
            for dict_ in list_ser:
                list_counter_dicts.append(Counter(dict_))
            return dict(sum(list_counter_dicts))

    def filter_list_ser(
        self,
        list_ser: List[Dict[str, Any]],
        foreigner_key: str,
        k_value: Number,
        str_filter_type: str = "equal",
    ) -> List[Dict[str, Any]]:
        """
        Filter list of dictionaries
        Args:
            list_ser (List[Dict[str, Any]]): List of dictionaries to filter
            foreigner_key (str): Foreigner key to filter
            k_value (Number): Value of foreigner key of interest
            str_filter_type (str): Type of filter to apply
        Returns:
            List[Dict[str, Any]]: Filtered list of dictionaries
        """
        if str_filter_type == "equal":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] == k_value]
        elif str_filter_type == "not_equal":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] != k_value]
        elif str_filter_type == "less_than":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] < k_value]
        elif str_filter_type == "greater_than":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] > k_value]
        elif str_filter_type == "less_than_or_equal_to":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] <= k_value]
        elif str_filter_type == "greater_than_or_equal_to":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] >= k_value]
        elif str_filter_type == "isin":
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] in k_value]
        else:
            raise ValueError(
                'str_filter_type must be "equal", "not_equal", "less_than" or "greater_than"'
            )

    def merge_values_foreigner_keys(
        self, list_ser, foreigner_key, list_keys_merge_dict
    ):
        """
        REFERECES: https://stackoverflow.com/questions/50167565/python-how-to-merge-dict-in-list-of-dicts-based-on-value
        DOCSTRING: MERGE DICTS ACCORDINGLY TO A FOREIGNER KEY IN THE LIST OF DICTS
        INPUTS: LIST OF DICTS, FOREIGNER KEY AND LIST OF KEYS TO MERGE IN A GIVEN SET O DICTS
        OUTPUTS: LIST OF DICTS
        """
        # setting default variables
        list_ser_export = list()
        list_foreinger_keys = list()
        list_ser_export = list()
        # get values from foreinger key
        list_foreinger_keys = list(set([dict_[foreigner_key] for dict_ in list_ser]))
        # iterating through list of foreigner key and merging values of interest
        for key in list_foreinger_keys:
            # filter dicts for the given foreinger key
            list_filtered_dicts = self.filter_list_ser(list_ser, foreigner_key, key)
            # merge dictionaries accordingly to given keys
            list_ser_export.append(
                self.sum_values_selected_keys(list_filtered_dicts, list_keys_merge_dict)
            )
        # return final result
        return list_ser_export

    def n_smallest(self, list_ser, key_, n):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return nsmallest(n, list_ser, key=lambda dict_: dict_[key_])

    def n_largest(self, list_ser, key_, n):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return nlargest(n, list_ser, key=lambda dict_: dict_[key_])

    def order_dict(self, dict_):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return OrderedDict(sorted(dict_.items()))

    def group_by_dicts(self, list_ser):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # sort by the desired field first
        list_ser.sort(key=itemgetter("date"))
        # return iteration in groups
        return groupby(list_ser, key=itemgetter("date"))

    def add_key_value_to_dicts(
        self,
        list_ser: List[Dict[str, Union[int, float, str]]],
        key: Union[str, List[Dict[str, Union[int, float, str]]]],
        value: Union[
            Callable[..., Union[int, float, str]], Union[int, float, str], None
        ] = None,
        list_keys_for_function: Optional[List[str]] = None,
        kwargs_static: Optional[Dict[str, Union[int, float, str, None]]] = None,
    ) -> List[Dict[str, Union[int, float, str]]]:
        """
        Adds a key and value (or multiple key-value pairs) to every dictionary in a list.

        Args:
            list_ser (List[Dict[str, Union[int, float, str]]]): A list of dictionaries to be updated.
            key (Union[str, List[Dict[str, Union[int, float, str]]]]):
                - If a string, the key to add to each dictionary
                - If a list of dictionaries, each dictionary contains key-value pairs to add
            value (Union[Callable[..., Union[int, float, str]], Union[int, float, str], None]):
                - The value to associate with the key (if key is a string)
                - Can be a static value or a function to compute the value dynamically
                - Ignored if key is a list of dictionaries
            list_keys_for_function (Optional[List[str]]):
                - Keys to extract from the dictionary for the value function (if callable)
                - Required if value is a function
            kwargs_static (Optional[Dict[str, Union[int, float, str, None]]]):
                - Static keyword arguments to pass to the function (if callable)

        Returns:
            List[Dict[str, Union[int, float, str]]]: The updated list of dictionaries.

        Raises:
            ValueError: if key is a string and value is None.
            TypeError: if key is neither a string nor a list of dictionaries.
        """
        # if key is a string, treat it as a single key-value pair
        if isinstance(key, str):
            if value is None:
                raise ValueError("If key is a string, value must be provided.")
            for dict_ in list_ser:
                if isinstance(dict_, dict):
                    if callable(value):
                        args = (
                            [dict_.get(k) for k in list_keys_for_function]
                            if list_keys_for_function is not None
                            else []
                        )
                        if kwargs_static is not None:
                            dict_[key] = value(*args, **kwargs_static)
                        else:
                            dict_[key] = value(*args)
                    else:
                        dict_[key] = value
        # if key is a list of dictionaries, treat it as multiple key-value pairs
        elif isinstance(key, list):
            for dict_ in list_ser:
                if isinstance(dict_, dict):
                    for kv_pair in key:
                        if isinstance(kv_pair, dict):
                            for k, v in kv_pair.items():
                                dict_[k] = v
        else:
            raise TypeError("key must be a string or a list of dictionaries.")
        return list_ser

    def pair_keys_with_values(
        self, list_keys: List[str], list_lists: List[List[Any]]
    ) -> List[Dict[str, Any]]:
        """
        Pair each sublist with the list of keys and return a list of dictionaries
        Args:
            list_keys (List[str]): The list of keys to pair with the sublists
            list_lists (List[List[Any]]): The list of sublists to pair with the keys
        Returns:
            List[Dict[str, Any]]: A list of dictionaries, where each dictionary is a pairing of
                a sublist with the list of keys
        Raises:
            ValueError: If the length of a sublist does not match the length of list_keys
        """
        list_ser = []
        for sublist in list_lists:
            if len(sublist) != len(list_keys):
                raise ValueError(
                    f"The length of the sublist {sublist} does not match the length of list_keys."
                )
            # Pair keys with values from the sublist
            dict_entry = {key: value for key, value in zip(list_keys, sublist)}
            list_ser.append(dict_entry)
        return list_ser

    def pair_headers_with_data(
        self, list_headers: List[str], list_data: List[Any]
    ) -> List[Dict[str, Any]]:
        """
        DOCSTRING: PAIR HEADERS AND DATA AS KEYS AND VALUES IN A SERIALIZED LIST
            - FOR EXAMPLE, IF LIST_HEADERS IS ['NAME', 'AGE'] AND LIST_DATA IS
                ['JOHN', 25, 'ALICE', 30], THE FUNCTION WILL RETURN [{'NAME': 'JOHN', 'AGE': 25},
                {'NAME': 'ALICE', 'AGE': 30}]
        INPUTS: LIST HEADERS, LIST DATA
        OUTPUTS: LIST
        """
        # setting variables
        list_ser = list()
        # ensuring the list_data length is a multiple of list_headers length
        if len(list_data) % len(list_headers) != 0:
            raise ValueError(
                "The length of list_data is not a multiple of the length of list_headers."
            )
        # iterate over the list_data in chunks equal to the length of list_headers
        for i in range(0, len(list_data), len(list_headers)):
            # create a dictionary for each chunk
            entry = {
                list_headers[j]: list_data[i + j] for j in range(len(list_headers))
            }
            list_ser.append(entry)
        # returning list of dictionaries
        return list_ser

    def fill_placeholders(
        self, dict_base: Dict[str, Any], dict_replacer: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Replaces placeholders in the base dictionary with values from the replacement dictionary
        Args:
            dict_base (Dict[str, Any]): The dictionary containing placeholders
            dict_replacer (Dict[str, Any]): The dictionary providing replacement values
        Returns:
            Dict[str, Any]: The updated dictionary with placeholders replaced.
        """
        placeholder_pattern = re.compile(r"\{\{\s*(\w+)\s*\}\}")
        def replace_value(value):
            if isinstance(value, dict):
                return {k: replace_value(v) for k, v in value.items()}
            elif isinstance(value, str):
                return placeholder_pattern.sub(
                    lambda m: str(dict_replacer.get(m.group(1), m.group(0))), value
                )
            elif isinstance(value, list):
                return [replace_value(v) for v in value]
            return value
        return {key: replace_value(value) for key, value in dict_base.items()}



import pandas as pd
from datetime import datetime
from typing import Optional
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from stpstone._config.global_slots import YAML_IRSBR
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests


class IRSBR(ABCRequests):

    def __init__(
        self,
        bl_create_session:bool=False,
        bl_new_proxy:bool=False,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        session: Optional[Session] = None,
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_IRSBR,
            session=session,
            dt_ref=dt_ref,
            dict_headers={
                'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
                'Accept-Language': 'pt-BR,pt;q=0.9',
                'Connection': 'keep-alive',
                'Referer': 'https://arquivos.receitafederal.gov.br/dados/cnpj/dados_abertos_cnpj/2025-01/',
                'Sec-Fetch-Dest': 'document',
                'Sec-Fetch-Mode': 'navigate',
                'Sec-Fetch-Site': 'same-origin',
                'Sec-Fetch-User': '?1',
                'Upgrade-Insecure-Requests': '1',
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36',
                'sec-ch-ua': '"Not A(Brand";v="8", "Chromium";v="132", "Google Chrome";v="132"',
                'sec-ch-ua-mobile': '?0',
                'sec-ch-ua-platform': '"Windows"'
            },
            dict_payload=None,
            cls_db=cls_db,
            logger=logger,
            year_dt_ref = DatesBR().year_number(self.dt_ref),
            month_dt_ref = DatesBR().month_number(self.dt_ref, bl_month_mm=True)
        )

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        return None



import numpy as np
import pandas as pd
from typing import Union, List
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class Arrays(metaclass=TypeChecker):

    def to_array_matrice(self, array_data: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        if len(array_data) == 0:
            raise ValueError("Return array is empty.")
        return array_data if isinstance(array_data, np.ndarray) else (
            array_data.to_numpy() if isinstance(array_data, pd.DataFrame) else None
        )

    def to_array_vector(self, array_data: Union[np.ndarray, pd.Series, List[float]]) -> np.ndarray:
        if len(array_data) == 0:
            raise ValueError("Return array is empty.")
        return array_data if isinstance(array_data, np.ndarray) else (
            np.array(array_data) if isinstance(array_data, list)
            else (array_data.to_numpy() if isinstance(array_data, pd.Series) else None)
        )

    def to_array(self, array_data: Union[np.ndarray, pd.DataFrame, pd.Series, List[float]]) \
        -> np.ndarray:
        if isinstance(array_data, (np.ndarray, pd.DataFrame)):
            return self.to_array_matrice(array_data)
        elif isinstance(array_data, (pd.Series, list, np.ndarray)):
            return self.to_array_vector(array_data)
        else:
            raise ValueError(f"Unsupported data type: {type(array_data)}")



import numpy as np
import numpy_financial as npf
from typing import List, Tuple
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class FinancialMath(metaclass=TypeChecker):

    def compound_r(self, float_ytm: float, int_nper: int, int_compound_n: int) -> float:
        return float((1.0 + float_ytm) ** (float(int_nper) / float(int_compound_n))) - 1.0

    def simple_r(self, float_ytm: float, int_nper: int, int_compound_n: int) -> float:
        return float_ytm * float(int_nper) / float(int_compound_n)

    def pv(
        self,
        float_ytm: float,
        int_nper: int,
        float_fv: float,
        float_pmt: float = 0,
        str_capitalization: str = "compound",
        str_when: str = "end"
    ) -> float:
        if str_capitalization == "compound":
            return npf.pv(float_ytm, int_nper, float_pmt, float_fv, str_when)
        elif str_capitalization == "simple":
            return float_fv / (1.0 + self.simple_r(float_ytm, int_nper, 1))
        else:
            raise ValueError("str_capitalization must be 'compound' or 'simple'")

    def fv(
        self,
        float_ytm: float,
        int_nper: int,
        float_pv: float,
        float_pmt: float = 0,
        str_capitalization: str = "compound",
        str_when: str = "end"
    ) -> float:
        if str_capitalization == "compound":
            return npf.fv(float_ytm, int_nper, float_pmt, float_pv, str_when)
        elif str_capitalization == "simple":
            return float_pv * (1.0 + self.simple_r(float_ytm, int_nper, 1))
        else:
            raise ValueError("str_capitalization must be 'compound' or 'simple'")

    def irr(self, list_cfs: List[float]) -> float:
        """
        Internal Rate of Return: interest rate at which the net present value of a list of
        cash flows is zero. The IRR is often used to evaluate the performance of an investment or
        project.

        Ags:
            list_cfs: List[float]
                A list of cash flows. The list must have at least one positive and one negative value.

        Returns:
            The internal rate of return of the cash flows.

        Raises:
            ValueError
                - If the list of cash flows does not have at least one positive and one negative
                value.
        """
        if (not any(float_cf < 0 for float_cf in list_cfs)) \
            or (not any(float_cf > 0 for float_cf in list_cfs)):
            raise ValueError(
                "List of cash flows must have at least one positive and one negative value.")
        return npf.irr(list_cfs)

    def npv(self, float_ytm: float, list_cfs: List[float]) -> float:
        """
        Net Present Value: the difference between the present value of the cash inflows and the
        present value of the cash outflows. The NPV is often used to evaluate the performance of an
        investment or project.

        Args:
            float_ytm: float
                The discount rate to be used when calculating the net present value.
            list_cfs: List[float]
                A list of cash flows. The list must have at least one positive and one negative value.

        Returns:
            The net present value of the cash flows.

        Raises:
            ValueError
                - If the list of cash flows does not have at least one positive and one negative value.
        """
        return npf.npv(float_ytm, list_cfs)

    def pmt(
        self,
        float_ytm: float,
        int_nper: int,
        float_pv: float,
        float_fv: float = 0,
        str_when: str = "end"
    ) -> float:
        """
        Calculate the fixed monthly payment based on a constant interest rate and a constant
        payment schedule.

        Args:
            float_ytm: float
                The interest rate per period.
            int_nper: int
                The total number of periods (months) over which the loan is outstanding.
            float_pv: float
                The present value of the loan/annuity. The present value is the lump-sum amount that
                a series of future payments is worth right now.
            float_fv: float, optional
                The future value of the loan/annuity. The future value is the amount of money the
                series of payments will be worth after the last payment is made. If not given, the
                future value is set to 0.
            str_when: str, optional
                When payments are due (either 'begin' or 'end'). If not given, 'end' is used.

        Returns:
            The fixed monthly payment.
        """
        return npf.pmt(float_ytm, int_nper, float_pv, float_fv, str_when)

    def pv_cfs(
        self,
        list_cfs: List[float],
        float_ytm: float,
        str_capitalization: str = "compound",
        str_when: str = "end"
    ) -> Tuple[np.ndarray, np.ndarray]:
        array_nper = np.arange(1, len(list_cfs) + 1)
        array_discounted_cfs = np.array([
            self.pv(float_ytm, int_t, float_cf, 0, str_capitalization, str_when)
            for int_t, float_cf in tuple(zip(array_nper, list_cfs))
        ])
        return array_nper, array_discounted_cfs



from typing import Union, Dict, List, Optional
from logging import Logger
from selenium.common.exceptions import TimeoutException
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD
from stpstone.utils.connections.netops.proxies.abc import ABCSession
from stpstone.utils.geography.ww import WWTimezones, WWGeography


class FreeProxyWorld(ABCSession):

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None,
        int_wait_load_seconds: int = 10,
    ) -> None:
        super().__init__(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )
        self.int_wait_load_seconds = int_wait_load_seconds
        self.fstr_url = "https://www.freeproxy.world/?type=&anonymity=&country={}&speed=&port=&page={}"
        self.xpath_tr = '//table[@class="layui-table"]//tbody/tr'

    @property
    def _available_proxies(self) -> List[Dict[str, Union[str, float]]]:
        list_ser = list()
        int_pg = 1
        while True:
            cls_selenium_wd = SeleniumWD(
                url=self.fstr_url.format(self.str_country_code.upper(), int_pg),
                bl_headless=True,
                bl_incognito=True,
                int_wait_load_seconds=self.int_wait_load_seconds
            )
            try:
                web_driver = cls_selenium_wd.get_web_driver
                try:
                    cls_selenium_wd.wait_until_el_loaded(self.xpath_tr)
                except TimeoutException:
                    self.create_log.log_message(
                        self.logger,
                        f"TimeoutException - URL: {self.fstr_url.format(self.str_country_code.upper(), int_pg)}",
                        "warning"
                    )
                    break
                el_trs = cls_selenium_wd.find_elements(web_driver, self.xpath_tr)
                list_range = [i for i in range(2, len(el_trs) + 2, 2)]
                for i_tr in list_range:
                    ip = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[1]').text
                    port = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[2]/a').text
                    country = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[3]//span[@class="table-country"]'
                                                    ).text
                    city = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[4]').text
                    speed = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]//div[@class="n-bar-wrapper"]/p/a'
                                                    ).text
                    type_ = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[6]/a').text
                    anonymity = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[7]/a').text.lower()
                    anonymity = "elite" if anonymity == "high" else "transparent"
                    last_checked = cls_selenium_wd.find_element(web_driver, self.xpath_tr \
                                                    + f'[{i_tr}]/td[8]').text
                    list_ser.append({
                        "protocol": type_.lower(),
                        "bl_alive": True,
                        "status": "success",
                        "alive_since": self.composed_time_ago_to_ts_unix(last_checked),
                        "anonymity": anonymity.lower(),
                        "average_timeout": 1.0 / self.proxy_speed_to_float(speed),
                        "first_seen": self.composed_time_ago_to_ts_unix(last_checked),
                        "ip_data": "",
                        "ip_name": "",
                        "timezone": ", ".join(WWTimezones().get_timezones_by_country_code(
                            self.str_country_code)),
                        "continent": WWGeography().get_continent_by_country_code(self.str_country_code),
                        "continent_code": WWGeography().get_continent_code_by_country_code(
                            self.str_country_code).upper(),
                        "country": country,
                        "country_code": self.str_country_code.upper(),
                        "city": city,
                        "district": "",
                        "region_name": "",
                        "zip": "",
                        "bl_hosting": True,
                        "isp": "",
                        "latitude": 0.0,
                        "longitude": 0.0,
                        "organization": "",
                        "proxy": True,
                        "ip": ip,
                        "port": port,
                        "bl_ssl": True,
                        "timeout": 1.0 / self.proxy_speed_to_float(speed),
                        "times_alive": 1,
                        "times_dead": "",
                        "ratio_times_alive_dead": 1.0,
                        "uptime": 1.0
                    })
                int_pg += 1
                cls_selenium_wd.wait(60)
            finally:
                web_driver.quit()
        return list_ser



import backoff
import chardet
import fitz
import pdfplumber
import urllib3
import os
import re
import subprocess
import tempfile
import pandas as pd
from time import sleep
from abc import ABC, abstractmethod
from datetime import datetime
from io import BytesIO, StringIO, TextIOWrapper
from logging import Logger
from typing import Any, Dict, List, Optional, Tuple, Type, Union, Literal
from urllib.parse import parse_qs, urlparse
from zipfile import ZipExtFile, ZipFile
from selenium.webdriver.remote.webdriver import WebDriver
from sqlalchemy.orm import Session
from requests import Request, Response, Session as ReqSession, request
from requests.exceptions import (ReadTimeout, ConnectTimeout, ChunkedEncodingError,
                                 RequestException, HTTPError, JSONDecodeError)
from stpstone.transformations.standardization.dataframe import DFStandardization
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.loggs.db_logs import DBLogs
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.folders import DirFilesManagement, RemoteFiles
from stpstone.utils.parsers.json import JsonFiles
from stpstone.utils.parsers.lists import ListHandler
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.xml import XMLFiles
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class UtilsRequests(ABC):

    def mock_response(
        self, file: Union[BytesIO, TextIOWrapper, StringIO, ZipExtFile], file_name: str
    ) -> Response:
        mock_resp = Response()
        mock_resp._content = file.read()
        mock_resp.url = file_name
        mock_resp.status_code = 200
        return mock_resp

    def columns_length(self, list_lines: List[Any]) -> int:
        dict_columns_count = {}
        for line in list_lines:
            num_columns = len(line.split(";"))
            dict_columns_count[num_columns] = dict_columns_count.get(num_columns, 0) + 1
        return max(dict_columns_count, key=dict_columns_count.get)

    def xml_find(
        self,
        soup_content: Type[XMLFiles],
        tag: str,
        tag_name: str,
        dict_xml_keys: Dict[str, Any],
    ) -> Dict[str, Union[str, float, int]]:
        dict_ = dict()
        soup_tag = soup_content.find(tag)
        if soup_tag is None:
            dict_[tag_name] = None
        else:
            dict_[tag_name] = soup_tag.get_text()
        if "attributes" in dict_xml_keys:
            for key_attrb_xml in dict_xml_keys["attributes"]:
                if (
                    (soup_tag is not None)
                    and (dict_xml_keys["attributes"][key_attrb_xml] is not None)
                    and (dict_xml_keys["attributes"][key_attrb_xml] in soup_tag.attrs)
                ):
                    dict_[dict_xml_keys["attributes"][key_attrb_xml]] = soup_tag.attrs[
                        dict_xml_keys["attributes"][key_attrb_xml]
                    ]
        return dict_

    def read_csv_with_error(
        self, file: Union[BytesIO, TextIOWrapper, StringIO, ZipExtFile]
    ) -> Union[BytesIO, TextIOWrapper, StringIO]:
        """
        Reads a file (BytesIO, TextIOWrapper, StringIO, ZipExtFile), corrects problematic lines,
        and returns the corrected content in the same file type as the input

        Args:
            file (Union[BytesIO, TextIOWrapper, StringIO, ZipExtFile]): The file to read.

        Returns:
            Union[BytesIO, TextIOWrapper, StringIO]: The corrected content in the same file type.
        """
        # reset the pointer to the beginning
        file.seek(0)
        if isinstance(file, BytesIO):
            content = file.read().decode("utf-8")
            list_lines = content.splitlines()
        elif isinstance(file, (TextIOWrapper, StringIO)):
            list_lines = file.readlines()
        elif isinstance(file, ZipExtFile):
            list_lines = [line.decode("utf-8") for line in file.readlines()]
        else:
            raise ValueError(
                "Unsupported file type. Expected BytesIO or TextIOWrapper."
            )
        list_corrected_lines = [
            line
            for line in list_lines
            if len(line.split(";")) == self.columns_length(list_lines)
        ]
        corrected_content = "\n".join(list_corrected_lines)
        if isinstance(file, BytesIO):
            return BytesIO(corrected_content.encode("utf-8"))
        elif isinstance(file, TextIOWrapper):
            return TextIOWrapper(BytesIO(corrected_content.encode("utf-8")))
        elif isinstance(file, StringIO):
            return StringIO(corrected_content)
        elif isinstance(file, ZipExtFile):
            return BytesIO(corrected_content.encode("utf-8"))

    def pdf_doc_tables_response(self, bytes_pdf: BytesIO) -> pd.DataFrame:
        list_ser = list()
        with pdfplumber.open(bytes_pdf) as pdf:
            for page in pdf.pages:
                list_ = page.extract_tables()
                list_ser.extend(
                    HandlingDicts().pair_keys_with_values(list_[0][0], list_[0][1:])
                )
        return pd.DataFrame(list_ser)

    def pdf_doc_regex(
        self,
        url: str,
        bytes_pdf: BytesIO,
        dict_regex_patterns: Dict[str, Dict[str, Any]],
        int_pgs_join: int = 2,
    ):
        list_pages = list()
        list_matches = list()
        dict_count_matches = dict()
        doc_pdf = fitz.open(
            stream=bytes_pdf,
            filetype=DirFilesManagement().get_last_file_extension(url),
        )
        str_ = ""
        for i in range(0, len(doc_pdf)):
            str_ = str_ + "\n" + doc_pdf[i].get_text("text")
            if (i % int_pgs_join == 0) \
                or (i == len(doc_pdf) - 1):
                list_pages.append(str_)
                str_ = ""
        for i, str_page in enumerate(list_pages):
            str_page = StrHandler().remove_diacritics_nfkd(str_page, bl_lower_case=True)
            if (len(dict_count_matches) > 0) \
                and (all(count > 0 for count in dict_count_matches.values())): break
            for str_event, dict_l1 in dict_regex_patterns.items():
                for str_condition, regex_pattern in dict_l1.items():
                    if str_event not in dict_count_matches:
                        dict_count_matches[str_event] = 0
                    if dict_count_matches[str_event] > 0:
                        break
                    regex_pattern = StrHandler().remove_diacritics_nfkd(
                        regex_pattern, bl_lower_case=False
                    )
                    regex_match = re.search(
                        regex_pattern,
                        str_page,
                        # flags=re.DOTALL | re.MULTILINE
                    )
                    if (regex_match is not None) \
                        and (regex_match.group(0) is not None) \
                        and (len(regex_match.group(0)) > 0):
                        dict_count_matches[str_event] += 1
                        dict_ = {
                            "EVENT": str_event.upper(),
                            "MATCH_PATTERN": str_condition.upper(),
                            "PATTERN_REGEX": regex_pattern,
                        }
                        for i_match in range(0, len(regex_match.groups()) + 1):
                            regex_group = regex_match.group(i_match).replace("\n", " ")
                            regex_group = re.sub(r"\s+", " ", regex_group).strip()
                            dict_[f"REGEX_GROUP_{i_match}"] = regex_group.upper()
                        list_matches.append(dict_)
                if dict_count_matches[str_event] == 0:
                    list_matches.append({
                        "EVENT": str_event.upper(),
                        "MATCH_PATTERN": "zzN/A",
                        "PATTERN_REGEX": "zzN/A",
                    })
        df_ = pd.DataFrame(list_matches)
        df_.drop_duplicates(inplace=True)
        df_.sort_values(by=["EVENT", "MATCH_PATTERN"], ascending=[True, True], inplace=True)
        df_.drop_duplicates(subset=["EVENT"], inplace=True)
        return df_
    
    def pivot_event_data(self, df_: pd.DataFrame) -> pd.DataFrame:
        """
        Pivots event data from long format to wide format with events as columns.
        
        Args:
            df_ (pd.DataFrame): Input DataFrame containing EVENT, REGEX_GROUP_1 columns
            
        Returns:
            pd.DataFrame: Pivoted DataFrame with one row per country and events as columns
        """
        df_['country_group'] = (df_['EVENT'] == 'COUNTRY_NAME').cumsum()
        df_pivoted = df_.pivot_table(
            index='country_group',
            columns='EVENT',
            values='REGEX_GROUP_1',
            aggfunc='first'
        ).reset_index(drop=True)
        df_pivoted.columns.name = None
        return df_pivoted

    def html_regex(
        self,
        html_content: str,
        dict_regex_patterns: Dict[str, Dict[str, Any]],
    ):
        dict_count_matches = dict()
        list_matches = list()
        with open("data/html_content.html", 'w', encoding='utf-8') as f:
            f.write(html_content)
        for str_event, dict_l1 in dict_regex_patterns.items():
            for str_condition, regex_pattern in dict_l1.items():
                if str_event not in dict_count_matches:
                    dict_count_matches[str_event] = 0
                regex_pattern = StrHandler().remove_diacritics_nfkd(
                    regex_pattern, bl_lower_case=False
                )
                regex_matches = list(re.finditer(regex_pattern, html_content, re.MULTILINE))
                if regex_matches:
                    for int_regex, match in enumerate(regex_matches):
                        dict_ = {
                            "INT_REGEX": int_regex,
                            "EVENT": str_event.upper(),
                            "MATCH_PATTERN": str_condition.upper(),
                            "PATTERN_REGEX": regex_pattern,
                            "REGEX_GROUP_0": match.group(0)
                        }
                        for i, group in enumerate(match.groups(), start=1):
                            dict_[f"REGEX_GROUP_{i}"] = group
                        list_matches.append(dict_)
                        dict_count_matches[str_event] += 1
            if dict_count_matches[str_event] == 0:
                list_matches.append({
                    "INT_REGEX": -1,
                    "EVENT": str_event.upper(),
                    "MATCH_PATTERN": "zzN/A",
                    "PATTERN_REGEX": "zzN/A",
                })
        df_ = pd.DataFrame(list_matches)
        df_.drop_duplicates(inplace=True)
        df_.sort_values(by=["INT_REGEX", "EVENT", "MATCH_PATTERN"], 
                        ascending=[True, True, True], inplace=True)
        return self.pivot_event_data(df_)

    def get_query_params(self, url: str, param: str) -> Union[int, bool, str, float, None]:
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.fragment)
        param_value = query_params.get(param, [None])[0]
        if param_value is not None:
            if isinstance(param_value, bool):
                return param_value
            elif param_value.lower() == 'true':
                return True
            elif param_value.lower() == 'false':
                return False
            try:
                return int(param_value)
            except ValueError:
                return param_value
        return None


class HandleReqResponses(UtilsRequests):

    def handle_response(
        self,
        resp_req: Union[Response, WebDriver],
        dict_xml_keys: Optional[Dict[str, Any]] = None,
        dict_regex_patterns: Optional[Dict[str, Dict[str, str]]] = None,
        dict_df_read_params: Optional[Dict[str, Any]] = None,
        dict_xpaths: Optional[Dict[str, str]] = None,
        list_ignored_file_extensions: Optional[List[str]] = [],
        list_ser_fixed_width_layout: Optional[List[Dict[str, Any]]] = [{}],
        selenium_wd: SeleniumWD = None
    ) -> pd.DataFrame:
        if isinstance(resp_req, WebDriver):
            url = resp_req.current_url
        else:
            url = resp_req.url
        str_file_extension = DirFilesManagement().get_last_file_extension(url)
        file_name = os.path.basename(url) if hasattr(resp_req, "url") else ""
        if self.req_trt_injection(resp_req) is not None:
            df_ = self.req_trt_injection(resp_req)
        elif str_file_extension == "xl_url":
            df_ = self._handle_pandas_excel_url(resp_req.url, dict_df_read_params)
        elif str_file_extension == "csv_url":
            df_ = self._handle_pandas_csv_url(resp_req.url, dict_df_read_params)
        elif isinstance(resp_req, WebDriver):
            df_ = self._handle_web_driver_html(url, resp_req, dict_xpaths, selenium_wd)
        elif str_file_extension == "zip":
            df_ = self._handle_zip_response(
                resp_req,
                dict_xml_keys,
                dict_regex_patterns,
                dict_df_read_params,
                list_ignored_file_extensions,
                list_ser_fixed_width_layout,
                selenium_wd
            )
        elif str_file_extension == "ex_":
            df_ = self._handle_ex_response(
                resp_req,
                dict_xml_keys,
                dict_regex_patterns,
                dict_df_read_params,
                list_ignored_file_extensions,
                list_ser_fixed_width_layout,
                selenium_wd
            )
        elif str_file_extension in ["csv", "txt", "asp", "do", "tex"]:
            if StrHandler().match_string_like(url, "*bl_separator_consistency_check=*") == True:
                bl_separator_consistency_check = (
                    self.get_query_params(url, "bl_separator_consistency_check")
                    if self.get_query_params(url, "bl_separator_consistency_check") is not None
                    else True
                )
            else:
                bl_separator_consistency_check = True
            if dict_df_read_params.get("encoding") is not None:
                resp_req.encoding = dict_df_read_params.get("encoding")
            if (bl_separator_consistency_check == True) \
                and (RemoteFiles().check_separator_consistency(
                    resp_req.content,
                    dict_df_read_params.get("skiprows", 0),
                    dict_df_read_params.get("skipfooter", 0)
                ) == False):
                df_ = self._handle_fwf_response(resp_req, list_ser_fixed_width_layout)
            else:
                df_ = self._handle_csv_response(resp_req, dict_df_read_params)
        elif str_file_extension in ["xlsx", "xls"]:
            df_ = self._handle_excel_response(resp_req, dict_df_read_params)
        elif str_file_extension == "xml":
            df_ = self._handle_xml_response(resp_req, dict_xml_keys)
        elif str_file_extension == "json":
            df_ = self._handle_json_response(resp_req)
        elif str_file_extension in ["pdf", "docx", "doc", "docm", "dot", "dotm"]:
            df_ = self._handle_pdf_doc_response(url, resp_req, dict_regex_patterns)
        elif str_file_extension in ["fwf", "dat"]:
            df_ = self._handle_fwf_response(resp_req, list_ser_fixed_width_layout)
        elif str_file_extension == "html_regex":
            df_ = self.html_regex(resp_req.text, dict_regex_patterns)
        else:
            try:
                json_ = resp_req.json()
            except JSONDecodeError:
                raise Exception(
                    "File extension not expected in the handle response method: "
                    + f"{str_file_extension}, please revisit the if-statement"
                )
            if isinstance(json_, dict) == True:
                df_ = pd.DataFrame([json_])
            else:
                df_ = pd.DataFrame(json_)
            return pd.DataFrame(json_)
        if (file_name is not None) \
            and ("FILE_NAME" not in df_.columns) \
            and (df_.empty == False):
            df_["FILE_NAME"] = file_name
        return df_

    @abstractmethod
    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        return None

    @backoff.on_exception(
        backoff.expo,
        (RequestException, HTTPError, ReadTimeout, ConnectTimeout, ChunkedEncodingError),
        max_tries=20,
        base=2,
        factor=2,
        max_value=1200
    )
    def _handle_web_driver_html(
        self,
        url: str,
        resp_req: WebDriver,
        dict_xpaths: Dict[str, str],
        selenium_wd: SeleniumWD
    ) -> pd.DataFrame:
        try:
            if any(StrHandler().match_string_like(value, "*delay_next_s=*") == True
                for value in dict_xpaths.values()):
                int_delay_next_s = (
                    self.get_query_params(url, "delay_next_s")
                    if self.get_query_params(url, "delay_next_s") is not None
                    else None
                )
            if any(StrHandler().match_string_like(value, "*{{iter}}*") == True
                for value in dict_xpaths.values()):
                int_iter_min = (
                    self.get_query_params(url, "iter_min")
                    if self.get_query_params(url, "iter_min") is not None
                    else 0
                )
                int_iter_max = (
                    self.get_query_params(url, "iter_max")
                    if self.get_query_params(url, "iter_max") is not None
                    else 100
                )
                list_ser = [{
                    key: selenium_wd.find_element(resp_req,
                    StrHandler().fill_placeholders(str_xpath, {"iter": i})).text
                    for key, str_xpath in dict_xpaths.items()
                    for i in range(int_iter_min, int_iter_max)
                }]
            else:
                list_ser = [{
                    key: selenium_wd.find_element(resp_req, str_xpath).text
                    for key, str_xpath in dict_xpaths.items()
                }]
        finally:
            resp_req.quit()
            if int_delay_next_s is not None: sleep(int_delay_next_s)
        return pd.DataFrame(list_ser)

    def _handle_zip_response(
        self,
        resp_req: Response,
        dict_xml_keys: Optional[Dict[str, Any]] = None,
        dict_regex_patterns: Optional[Dict[str, Dict[str, str]]] = None,
        dict_df_read_params: Optional[Dict[str, Any]] = None,
        list_ignored_file_extensions: Optional[List[str]] = [],
        list_ser_fixed_width_layout: Optional[List[Dict[str, Any]]] = [{}],
        selenium_wd: SeleniumWD = None
    ) -> Union[pd.DataFrame, List[pd.DataFrame]]:
        zipfile = ZipFile(BytesIO(resp_req.content))
        list_ = []
        for file_name in zipfile.namelist():
            with zipfile.open(file_name) as file:
                str_file_extension = DirFilesManagement().get_last_file_extension(
                    file_name
                )
                if str_file_extension in list_ignored_file_extensions:
                    continue
                mock_resp = self.mock_response(file, file_name)
                df_ = self.handle_response(
                    mock_resp,
                    dict_xml_keys,
                    dict_regex_patterns,
                    dict_df_read_params,
                    list_ignored_file_extensions,
                    list_ser_fixed_width_layout,
                    selenium_wd
                )
                if df_.empty == True:
                    continue
                elif isinstance(df_, pd.DataFrame):
                    list_.extend(df_.to_dict(orient="records"))
                elif isinstance(df_, list):
                    for df_nested in df_:
                        list_.extend(df_nested.to_dict(orient="records"))
        if list_:
            return pd.DataFrame(list_)
        else:
            return pd.DataFrame()

    def _handle_csv_response(
        self,
        file: Union[BytesIO, TextIOWrapper, Response],
        dict_df_read_params: Optional[Dict[str, Any]],
    ) -> pd.DataFrame:
        if isinstance(file, BytesIO):
            file.seek(0)
        elif isinstance(file, Response):
            file = StringIO(file.text)
        try:
            return pd.read_csv(file, **dict_df_read_params)
        except pd.errors.ParserError:
            file = self.read_csv_with_error(file)
            if dict_df_read_params is not None:
                dict_df_read_corrected_params = dict_df_read_params.copy()
                dict_df_read_corrected_params.pop("skiprows", None)
            else:
                dict_df_read_corrected_params = dict()
            return pd.read_csv(file, **dict_df_read_corrected_params)

    def _handle_pandas_excel_url(
        self,
        url: Union[str],
        dict_df_read_params: Optional[Dict[str, Any]],
    ) -> pd.DataFrame:
        dict_df_read_params = dict_df_read_params if dict_df_read_params is not None else {}
        return pd.read_excel(url, **dict_df_read_params)

    def _handle_pandas_csv_url(
        self,
        url: Union[str],
        dict_df_read_params: Optional[Dict[str, Any]],
    ) -> pd.DataFrame:
        dict_df_read_params = dict_df_read_params if dict_df_read_params is not None else {}
        return pd.read_csv(url, **dict_df_read_params)

    def _handle_excel_response(
        self,
        file: Union[BytesIO, TextIOWrapper, Response],
        dict_df_read_params: Optional[Dict[str, Any]],
    ) -> pd.DataFrame:
        if isinstance(file, BytesIO):
            file.seek(0)
        elif isinstance(file, Response):
            file = StringIO(file.text)
        dict_df_read_params = dict_df_read_params if dict_df_read_params is not None else {}
        return pd.read_excel(file, **dict_df_read_params)

    def _handle_xml_response(
        self,
        file: Union[BytesIO, TextIOWrapper, Response],
        dict_xml_keys: Dict[str, Any],
    ) -> pd.DataFrame:
        list_ser = list()
        if isinstance(file, BytesIO):
            file.seek(0)
        if isinstance(file, Response):
            file = StringIO(file.text)
        soup_xml = XMLFiles().memory_parser(file)
        for key, list_tags in dict_xml_keys["tags"].items():
            for soup_content in soup_xml.find_all(key):
                for tag in list_tags:
                    if isinstance(tag, str):
                        dict_l1 = self.xml_find(soup_content, tag, tag, dict_xml_keys)
                        if ("dict_" in locals()) and (isinstance(dict_, dict)):
                            dict_ = HandlingDicts().merge_n_dicts(dict_, dict_l1)
                        else:
                            dict_ = dict_l1
                    elif isinstance(tag, dict):
                        key_ = list(tag.keys())[0]
                        list_values = list(tag.values())[0]
                        for soup_l2 in soup_content.find_all(key_):
                            for tag_l2 in list_values:
                                dict_l2 = self.xml_find(
                                    soup_l2, tag_l2, f"{key_}{tag_l2}", dict_xml_keys
                                )
                                if ("dict_" in locals()) and (isinstance(dict_, dict)):
                                    dict_ = HandlingDicts().merge_n_dicts(
                                        dict_, dict_l2
                                    )
                                else:
                                    dict_ = dict_l2
            dict_ = HandlingDicts().add_key_value_to_dicts(dict_, "ROOT_TAG", key)
            list_ser.append(dict_)
        return pd.DataFrame(list_ser)

    def _handle_json_response(
        self, file: Union[BytesIO, TextIOWrapper]
    ) -> pd.DataFrame:
        if isinstance(file, BytesIO):
            file.seek(0)
        json_file = file.read()
        list_ser = JsonFiles().loads_message_like(json_file)
        df_ = pd.DataFrame(list_ser)
        return df_

    def _handle_pdf_doc_response(
        self,
        url: str,
        resp_req: Response,
        dict_regex_patterns: Dict[str, Dict[str, str]],
        default_int_pgs_join: int = 2
    ) -> pd.DataFrame:
        bytes_pdf = BytesIO(resp_req.content)
        # checking wheter to read tables or use regex
        if StrHandler().match_string_like(url, "*feat=read_tables*") == True:
            return self.pdf_doc_tables_response(bytes_pdf)
        else:
            if StrHandler().match_string_like(url, "*int_pgs_join=*") == True:
                int_pgs_join = (
                    self.get_query_params(url, "int_pgs_join")
                    if self.get_query_params(url, "int_pgs_join") is not None
                    else default_int_pgs_join
                )
            else:
                int_pgs_join = default_int_pgs_join
            return self.pdf_doc_regex(
                url, bytes_pdf, dict_regex_patterns, int_pgs_join
            )

    def _handle_ex_response(
        self,
        resp_req: Response,
        dict_xml_keys: Optional[Dict[str, Any]] = None,
        dict_regex_patterns: Optional[Dict[str, Dict[str, str]]] = None,
        dict_df_read_params: Optional[Dict[str, Any]] = None,
        list_ignored_file_extensions: Optional[List[str]] = [],
        list_ser_fixed_width_layout: Optional[List[Dict[str, Any]]] = [{}],
        selenium_wd: SeleniumWD = None
    ) -> pd.DataFrame:
        list_ser = list()
        with tempfile.TemporaryDirectory() as temp_dir_path:
            ex_file_path = RemoteFiles().get_file_from_zip(
                resp_req, temp_dir_path, (".ex_")
            )
            os.chmod(ex_file_path, 0o755)
            subprocess.run([ex_file_path], cwd=temp_dir_path, check=True)
            list_files = [
                f
                for f in os.listdir(temp_dir_path)
                if f != os.path.basename(ex_file_path) and not f.endswith(".zip")
            ]
            for file_name in list_files:
                file_path = os.path.join(temp_dir_path, file_name)
                with open(file_path, "rb") as file:
                    mock_resp = self.mock_response(file, file_name)
                    df_ = self.handle_response(
                        mock_resp,
                        dict_xml_keys,
                        dict_regex_patterns,
                        dict_df_read_params,
                        list_ignored_file_extensions,
                        list_ser_fixed_width_layout,
                        selenium_wd
                    )
                    list_ser.extend(df_.to_dict(orient="records"))
            return pd.DataFrame(list_ser)

    def _handle_fwf_response(
        self, resp_req: Response, list_ser_fixed_width_layout: List[Dict[str, Any]]
    ) -> pd.DataFrame:
        list_colspecs = [
            (dict_["start"], dict_["end"]) for dict_ in list_ser_fixed_width_layout
        ]
        list_colnames = [dict_["field"] for dict_ in list_ser_fixed_width_layout]
        if (
            (resp_req.headers is not None)
            and ("application/zip" in resp_req.headers.get("Content-Type", ""))
            and (".zip" in resp_req.headers.get("Content-Disposition"))
        ):
            with tempfile.TemporaryDirectory() as temp_dir_path:
                file_path = RemoteFiles().get_file_from_zip(
                    resp_req, temp_dir_path, (".fwf", ".dat", ".txt", "")
                )
                return pd.read_fwf(
                    file_path, colspecs=list_colspecs, names=list_colnames
                )
        else:
            try:
                encoding = chardet.detect(resp_req.content)["encoding"]
            except:
                encoding = "latin-1"
            decoded_content = resp_req.content.decode(encoding)
            return pd.read_fwf(
                StringIO(decoded_content), colspecs=list_colspecs, names=list_colnames
            )


class ABCRequests(HandleReqResponses):

    def __init__(
        self,
        dict_metadata: Dict[str, Any],
        session: Optional[ReqSession] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None,
        path_webdriver: Optional[str] = None,
        int_port: Optional[int] = None,
        str_user_agent: Optional[str] = None,
        int_wait_load_seconds: int = 10,
        int_delay_seconds: int = 10,
        bl_headless: bool = False,
        bl_incognito: bool = False,
        bl_ts_log_str: bool = True
    ) -> None:
        self.dict_metadata = dict_metadata
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.list_slugs = list_slugs
        self.path_webdriver = path_webdriver
        self.int_port = int_port
        self.str_user_agent = str_user_agent
        self.int_wait_load_seconds = int_wait_load_seconds
        self.int_delay_seconds = int_delay_seconds
        self.bl_headless = bl_headless
        self.bl_incognito = bl_incognito
        self.bl_ts_log_str = bl_ts_log_str
        self.list_options_wd = None \
            if self.dict_metadata["credentials"].get("web_driver", None) is None \
            else self.dict_metadata["credentials"]["web_driver"]["options"]
        self.pattern_special_http_chars = r'["<>#%{}|\\^~\[\]` ]'
        self.token = (
            token
            if token is not None
            else (
                self.access_token
                if self.dict_metadata["credentials"]["token"]["host"] is not None
                else None
            )
        )
        self.create_log = CreateLog()

    @property
    def access_token(self):
        dict_instance_vars = self.get_instance_variables
        url_token = (
            self.dict_metadata["credentials"]["token"]["host"]
            + self.dict_metadata["credentials"]["token"]["app"]
            if self.dict_metadata["credentials"]["token"]["app"] is not None
            else self.dict_metadata["credentials"]["token"]["host"]
        )
        url_token = StrHandler().fill_placeholders(url_token, dict_instance_vars)
        resp_req = self.generic_req(
            self.dict_metadata["credentials"]["token"]["get"]["req_method"],
            url_token,
            self.dict_metadata["credentials"]["token"]["get"]["bl_verify"],
            self.dict_metadata["credentials"]["token"]["get"]["timeout"],
            self.dict_metadata["credentials"]["token"].get("headers", None),
        )
        return resp_req.json()[
            self.dict_metadata["credentials"]["token"]["keys"]["token"]
        ]

    @property
    def get_instance_variables(self) -> Dict[str, Any]:
        """
        Gather all instance variables (self.*) into a dictionary.
        """
        return {
            key: value
            for key, value in self.__dict__.items()
            if not key.startswith("__")
        }

    # ! TODO: implement timeout
    def generic_req_w_session(
        self,
        req_method: str,
        url: str,
        bl_verify: bool,
        tup_timeout: Tuple[float, float] = (12.0, 12.0),
        dict_headers: Optional[Dict[str, str]] = None,
        payload: Optional[Union[str, Dict[str, str]]] = None,
        cookies: Optional[Dict[str, str]] = None
    ) -> Response:
        if re.search(self.pattern_special_http_chars, url):
            #   prepare the request manually to preserve special characters
            resp_req = Request(req_method, url, headers=dict_headers, params=payload, cookies=cookies)
            req_preppped = self.session.prepare_request(resp_req)
            req_preppped.url = url
            resp_req = self.session.send(req_preppped, verify=bl_verify)
        else:
            if req_method == "GET":
                resp_req = self.session.get(
                    url, verify=bl_verify, headers=dict_headers, params=payload, cookies=cookies
                )
            elif req_method == "POST":
                resp_req = self.session.post(
                    url, verify=bl_verify, headers=dict_headers, data=payload, cookies=cookies
                )
        resp_req.raise_for_status()
        return resp_req

    # ! TODO: implement timeout
    @backoff.on_exception(
        backoff.expo,
        (RequestException, HTTPError, ReadTimeout, ConnectTimeout, ChunkedEncodingError),
        max_tries=20,
        base=2,
        factor=2,
        max_value=1200
    )
    def generic_req_wo_session(
        self,
        req_method: str,
        url: str,
        bl_verify: bool,
        tup_timeout: Tuple[float, float] = (12.0, 12.0),
        dict_headers: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None
    ) -> Response:
        if re.search(self.pattern_special_http_chars, url):
            #   prepare the request manually to preserve special characters
            req = Request(req_method, url, headers=dict_headers, params=payload, cookies=cookies)
            req_preppped = req.prepare()
            with ReqSession() as session:
                resp_req = session.send(req_preppped, verify=bl_verify)
        else:
            return request(
                req_method,
                url,
                verify=bl_verify,
                headers=dict_headers,
                params=payload if req_method == "GET" else None,
                data=payload if req_method == "POST" else None,
                cookies=cookies,
                # timeout=tup_timeout
            )
        resp_req.raise_for_status()
        return resp_req

    def generic_req(
        self,
        req_method: Literal["GET", "POST"],
        url: str,
        bl_verify: bool,
        tup_timeout: Tuple[float, float] = (12.0, 12.0),
        dict_headers: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None
    ) -> Response:
        """
        Check wheter the request method is valid and do the request with distinctions for session and
            local proxy-based requests
        Args:
            req_method (str): request method
            url (str): request url
            bl_verify (bool): verify request
            tup_timeout (Tuple[float, float]): request timeout
        Returns:
            Tuple[Response, str]
        """
        if self.session is not None:
            func_generic_req = self.generic_req_w_session
        else:
            func_generic_req = self.generic_req_wo_session
        resp_req = func_generic_req(
            req_method, url, bl_verify, tup_timeout, dict_headers, payload, cookies
        )
        return resp_req

    def trt_req(
        self,
        req_method: str,
        host: str,
        dict_dtypes: Dict[str, Any],
        dict_headers: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, str]] = None,
        app: Optional[str] = None,
        bl_verify: bool = False,
        tup_timeout: Tuple[float, float] = (12.0, 12.0),
        cookies: Optional[Dict[str, str]] = None,
        cols_from_case: Optional[str] = None,
        cols_to_case: Optional[str] = None,
        list_cols_drop_dupl: List[str] = None,
        str_fmt_dt: str = "YYYY-MM-DD",
        type_error_action: str = "raise",
        strategy_keep_when_dupl: str = "first",
        dict_regex_patterns: Optional[Dict[str, Dict[str, str]]] = None,
        dict_df_read_params: Optional[Dict[str, Any]] = None,
        list_ignored_file_extensions: Optional[List[str]] = [],
        list_ser_fixed_width_layout: Optional[List[Dict[str, Any]]] = [{}],
        dict_xml_keys: Optional[Dict[str, Any]] = None,
        dict_fillna_strt: Optional[Dict[str, str]] = {},
        xpath_el_wait_until_loaded: Optional[str] = None,
        dict_xpaths: Optional[Dict[str, str]] = None
    ) -> pd.DataFrame:
        if payload is not None: payload = JsonFiles().dict_to_json(payload)
        url = host + app if app is not None else host
        self.create_log.log_message(self.logger, f"Starting request to: {url}", "info")
        if self.dict_metadata["credentials"].get("web_driver", None) is not None:
            dict_instance_vars = self.get_instance_variables
            list_options_wd = [
                StrHandler().fill_placeholders(x, dict_instance_vars) for x in self.list_options_wd
            ]
            if self.session is not None:
                str_proxy = self.session.proxies
            else:
                str_proxy = None
            selenium_wd = SeleniumWD(
                url, self.path_webdriver, self.int_port, self.str_user_agent, self.int_wait_load_seconds,
                self.int_delay_seconds, str_proxy, self.bl_headless, self.bl_incognito,
                list_options_wd
            )
            if xpath_el_wait_until_loaded is not None:
                selenium_wd.wait_until_el_loaded(xpath_el_wait_until_loaded)
            resp_req = selenium_wd.web_driver
        else:
            selenium_wd = None
            resp_req = self.generic_req(
                req_method, url, bl_verify, tup_timeout, dict_headers, payload, cookies
            )
        df_ = self.handle_response(
            resp_req,
            dict_xml_keys,
            dict_regex_patterns,
            dict_df_read_params,
            dict_xpaths,
            list_ignored_file_extensions,
            list_ser_fixed_width_layout,
            selenium_wd
        )
        cls_df_stdz = DFStandardization(
            dict_dtypes=dict_dtypes,
            cols_from_case=cols_from_case,
            cols_to_case=cols_to_case,
            list_cols_drop_dupl=list_cols_drop_dupl,
            dict_fillna_strt=dict_fillna_strt,
            str_fmt_dt=str_fmt_dt,
            type_error_action=type_error_action,
            strategy_keep_when_dupl=strategy_keep_when_dupl,
            encoding=(
                dict_df_read_params.get("encoding", "latin-1")
                if dict_df_read_params is not None
                else "latin-1"
            ),
            logger=self.logger,
        )
        df_ = cls_df_stdz.pipeline(df_)
        df_ = DBLogs().audit_log(df_, url, self.dt_ref, self.bl_ts_log_str)
        return df_

    def insert_table(
        self,
        str_resource: str,
        list_ser: Optional[List[Dict[str, Any]]] = None,
        bl_insert_or_ignore: bool = False,
        bl_schema: bool = True,
    ) -> None:
        """
        Insert data into data table
        Args:
            str_resource (str): resource name
            list_ser (List[Dict[str, Any]]): data to insert
            bl_insert_or_ignore (bool): False as default
            bl_schema (bool): some databases, like SQLite, don't have schemas - True as default
        Raises:
            Exception: If database or data is not defined
        """
        if self.cls_db is None:
            raise Exception(
                "Data insertion failed due to lack of database definition. "
                "Please revisit this parameter."
            )
        if list_ser is None:
            raise Exception(
                "Data insertion failed due to lack of data. "
                "Please revisit this parameter."
            )
        if str_resource == "general":
            return None
        elif str_resource == "metadata":
            for _, dict_ in self.dict_metadata[str_resource].items():
                str_table_name = dict_["table_name"]
                if bl_schema == False:
                    str_table_name = f"{dict_['schema']}_{str_table_name}"
                self.cls_db.insert(
                    dict_["data"],
                    str_table_name=str_table_name,
                    bl_insert_or_ignore=True,
                )
        else:
            str_table_name = self.dict_metadata[str_resource]["table_name"]
            if bl_schema == False:
                str_table_name = f"{self.dict_metadata[str_resource]['schema']}_{str_table_name}"
            self.cls_db.insert(
                list_ser,
                str_table_name=str_table_name,
                bl_insert_or_ignore=bl_insert_or_ignore,
            )

    def non_iteratively_get_data(
        self,
        str_resource: str,
        host: Optional[str] = None,
        bl_fetch: bool = False,
    ) -> Optional[pd.DataFrame]:
        """
        Non-iteratively get raw data
        Args:
            str_resource (str): resource name
            host (Optional[str]): host name
            bl_fetch (bool): False as default
        Returns:
            pd.DataFrame
        """
        dict_instance_vars = self.get_instance_variables
        app_ = self.dict_metadata[str_resource].get("app", None)
        app_ = StrHandler().fill_placeholders(app_, dict_instance_vars)
        host_ = (
            host
            if host is not None
            else (
                self.dict_metadata[str_resource].get("host", None)
                if self.dict_metadata[str_resource].get("host", None) is not None
                else self.dict_metadata["credentials"]["host"]
            )
        )
        host_ = StrHandler().fill_placeholders(host_, dict_instance_vars)
        dict_headers = (
            self.dict_metadata[str_resource]["headers"]
            if self.dict_metadata[str_resource].get("headers", None) is not None
            else self.dict_metadata["credentials"].get("headers", None)
        )
        if dict_headers is not None:
            dict_headers = HandlingDicts().fill_placeholders(
                dict_headers, dict_instance_vars
            )
        payload = (
            self.dict_metadata[str_resource]["payload"]
            if self.dict_metadata[str_resource].get("payload", None) is not None
            else self.dict_metadata["credentials"].get("payload", None)
        )
        if payload is not None:
            if isinstance(payload, dict):
                payload = HandlingDicts().fill_placeholders(payload, dict_instance_vars)
            elif isinstance(payload, str):
                payload = StrHandler().fill_placeholders(payload, dict_instance_vars)
            else:
                raise Exception("Payload must be either dict or str.")
        list_ignorable_exceptions = (
            self.dict_metadata[str_resource].get("list_ignorable_exceptions", list())
            if self.dict_metadata[str_resource].get("list_ignorable_exceptions", list())
            is not None
            else list()
        )
        list_ignorable_exceptions = [
            eval(exception) if isinstance(exception, str) else exception
            for exception in list_ignorable_exceptions
        ]
        # requiring data
        try:
            df_ = self.trt_req(
                self.dict_metadata[str_resource]["req_method"],
                host_,
                self.dict_metadata[str_resource]["dtypes"],
                dict_headers,
                payload,
                app_,
                self.dict_metadata[str_resource]["bl_verify"],
                self.dict_metadata[str_resource]["timeout"],
                self.dict_metadata[str_resource].get("cookies", None),
                self.dict_metadata[str_resource]["cols_from_case"],
                self.dict_metadata[str_resource]["cols_to_case"],
                self.dict_metadata[str_resource]["list_cols_drop_dupl"],
                self.dict_metadata[str_resource]["str_fmt_dt"],
                self.dict_metadata[str_resource]["type_error_action"],
                self.dict_metadata[str_resource]["strategy_keep_when_dupl"],
                self.dict_metadata[str_resource].get("regex_patterns", None),
                self.dict_metadata[str_resource].get("df_read_params", None),
                self.dict_metadata[str_resource].get("ignored_file_extensions", []),
                self.dict_metadata[str_resource].get("fixed_width_layout", [{}]),
                self.dict_metadata[str_resource].get("xml_keys", None),
                self.dict_metadata[str_resource].get("fillna_strt", {}),
                self.dict_metadata[str_resource]["web_driver"].get("xpath_el_wait_until_loaded", {}),
                self.dict_metadata[str_resource].get("xpaths", [{}]),
            )
        except tuple(list_ignorable_exceptions) as e:
            self.create_log.log_message(
                self.logger,
                "Iteration encountered an ignorable exception "
                + f"{e.__class__.__name__}: {e}. Continuing...",
                "warning"
            )
        except Exception as e:
            self.create_log.log_message(
                self.logger,
                f"Iteration failed due to {e.__class__.__name__}: {e}",
                "critical"
            )
            raise Exception(
                f"Iteration failed due to {e.__class__.__name__}: {e}"
            )
        if self.cls_db is not None:
            self.insert_table(
                str_resource,
                df_.to_dict(orient="records"),
                self.dict_metadata[str_resource]["bl_insert_or_ignore"],
                self.dict_metadata[str_resource]["bl_schema"],
            )
        if bl_fetch == True:
            return df_
        else:
            return None

    def iteratively_get_data(
        self,
        str_resource: str,
        host: Optional[str] = None,
        bl_fetch: bool = False,
    ) -> Optional[pd.DataFrame]:
        """
        Iteratively get raw data
        Args:
            str_resource (str): resource name
            host (Optional[str]): host name
            bl_fetch (bool): False as default
        Returns:
            Optional[pd.DataFrame]
        """
        list_ser = list()
        i = 0
        dict_instance_vars = self.get_instance_variables
        app_ = self.dict_metadata[str_resource].get("app", None)
        app_ = StrHandler().fill_placeholders(app_, dict_instance_vars)
        host_ = (
            host
            if host is not None
            else (
                self.dict_metadata[str_resource].get("host", None)
                if self.dict_metadata[str_resource].get("host", None) is not None
                else self.dict_metadata["credentials"]["host"]
            )
        )
        host_ = StrHandler().fill_placeholders(host_, dict_instance_vars)
        dict_headers = (
            self.dict_metadata[str_resource]["headers"]
            if self.dict_metadata[str_resource].get("headers", None) is not None
            else self.dict_metadata["credentials"].get("headers", None)
        )
        if dict_headers is not None:
            dict_headers = HandlingDicts().fill_placeholders(
                dict_headers, dict_instance_vars
            )
        payload = (
            self.dict_metadata[str_resource]["payload"]
            if self.dict_metadata[str_resource].get("payload", None) is not None
            else self.dict_metadata["credentials"].get("payload", None)
        )
        if payload is not None:
            if isinstance(payload, dict):
                payload = HandlingDicts().fill_placeholders(payload, dict_instance_vars)
            elif isinstance(payload, str):
                payload = StrHandler().fill_placeholders(payload, dict_instance_vars)
            else:
                raise Exception("Payload must be either dict or str.")
        list_slugs = (
            self.list_slugs
            if self.list_slugs is not None
            else self.dict_metadata[str_resource].get("slugs", None)
        )
        list_ignorable_exceptions = (
            self.dict_metadata[str_resource].get("list_ignorable_exceptions", list())
            if self.dict_metadata[str_resource].get("list_ignorable_exceptions", list())
            is not None
            else list()
        )
        list_ignorable_exceptions = [
            eval(exception) if isinstance(exception, str) else exception
            for exception in list_ignorable_exceptions
        ]
        # iterating through slugs/number of pages
        if list_slugs is not None:
            str_extract_from_braces = (
                StrHandler().extract_info_between_braces(app_)[0]
                if StrHandler().extract_info_between_braces(app_) is not None
                else ""
            )
            if str_extract_from_braces == "slug":
                for str_slug in list_slugs:
                    try:
                        df_ = self.trt_req(
                            self.dict_metadata[str_resource]["req_method"],
                            host_,
                            self.dict_metadata[str_resource]["dtypes"],
                            dict_headers,
                            payload,
                            StrHandler().fill_placeholders(app_, {"slug": str_slug}),
                            self.dict_metadata[str_resource]["bl_verify"],
                            self.dict_metadata[str_resource]["timeout"],
                            self.dict_metadata[str_resource].get("cookies", None),
                            self.dict_metadata[str_resource]["cols_from_case"],
                            self.dict_metadata[str_resource]["cols_to_case"],
                            self.dict_metadata[str_resource]["list_cols_drop_dupl"],
                            self.dict_metadata[str_resource]["str_fmt_dt"],
                            self.dict_metadata[str_resource]["type_error_action"],
                            self.dict_metadata[str_resource]["strategy_keep_when_dupl"],
                            self.dict_metadata[str_resource].get("regex_patterns", None),
                            self.dict_metadata[str_resource].get("df_read_params", None),
                            self.dict_metadata[str_resource].get("ignored_file_extensions", []),
                            self.dict_metadata[str_resource].get("fixed_width_layout", [{}]),
                            self.dict_metadata[str_resource].get("xml_keys", None),
                            self.dict_metadata[str_resource].get("fillna_strt", {}),
                            self.dict_metadata[str_resource].get("web_driver", {}).get(
                                "xpath_el_wait_until_loaded", {}),
                            self.dict_metadata[str_resource].get("xpaths", [{}]),
                        )
                        df_["SLUG_URL"] = str_slug
                        if df_.empty == True: break
                        list_ser.extend(df_.to_dict(orient="records"))
                        if self.cls_db is not None:
                            self.insert_table(
                                str_resource,
                                list_ser,
                                self.dict_metadata[str_resource]["bl_insert_or_ignore"],
                                self.dict_metadata[str_resource]["bl_schema"],
                            )
                            if bl_fetch == False:
                                list_ser = list()
                    except tuple(list_ignorable_exceptions) as e:
                        self.create_log.log_message(
                            self.logger,
                            "Iteration encountered an ignorable exception "
                            + f"{e.__class__.__name__}: {e}. Continuing...",
                            "warning"
                        )
                    except Exception as e:
                        self.create_log.log_message(
                            self.logger,
                            f"Iteration failed due to {e.__class__.__name__}: {e}",
                            "critical"
                        )
                        raise Exception(
                            f"Iteration failed due to {e.__class__.__name__}: {e}"
                        )
                    sleep(self.int_delay_seconds)
            elif str_extract_from_braces == "chunk_slugs":
                list_chunks_slugs = ListHandler().chunk_list(
                    list_to_chunk=self.list_slugs,
                    str_character_divides_clients=",",
                    int_chunk=self.dict_metadata[str_resource].get(
                        "int_chunk_slugs", 10
                    ),
                )
                for str_chunk_slugs in list_chunks_slugs:
                    try:
                        df_ = self.trt_req(
                            self.dict_metadata[str_resource]["req_method"],
                            host_,
                            self.dict_metadata[str_resource]["dtypes"],
                            dict_headers,
                            payload,
                            StrHandler().fill_placeholders(
                                app_, {"chunk_slugs": str_chunk_slugs}
                            ),
                            self.dict_metadata[str_resource]["bl_verify"],
                            self.dict_metadata[str_resource]["timeout"],
                            self.dict_metadata[str_resource].get("cookies", None),
                            self.dict_metadata[str_resource]["cols_from_case"],
                            self.dict_metadata[str_resource]["cols_to_case"],
                            self.dict_metadata[str_resource]["list_cols_drop_dupl"],
                            self.dict_metadata[str_resource]["str_fmt_dt"],
                            self.dict_metadata[str_resource]["type_error_action"],
                            self.dict_metadata[str_resource]["strategy_keep_when_dupl"],
                            self.dict_metadata[str_resource].get("regex_patterns", None),
                            self.dict_metadata[str_resource].get("df_read_params", None),
                            self.dict_metadata[str_resource].get("ignored_file_extensions", []),
                            self.dict_metadata[str_resource].get("fixed_width_layout", [{}]),
                            self.dict_metadata[str_resource].get("xml_keys", None),
                            self.dict_metadata[str_resource].get("fillna_strt", {}),
                            self.dict_metadata[str_resource]["web_driver"].get(
                                "xpath_el_wait_until_loaded", {}),
                            self.dict_metadata[str_resource].get("xpaths", [{}]),
                        )
                        df_["SLUG_URL"] = str_chunk_slugs
                        list_ser.extend(df_.to_dict(orient="records"))
                        if self.cls_db is not None:
                            self.insert_table(
                                str_resource,
                                list_ser,
                                self.dict_metadata[str_resource]["bl_insert_or_ignore"],
                                self.dict_metadata[str_resource]["bl_schema"],
                            )
                            if bl_fetch == False:
                                list_ser = list()
                    except tuple(list_ignorable_exceptions) as e:
                        self.create_log.log_message(
                            self.logger,
                            "Iteration encountered an ignorable exception "
                            + f"{e.__class__.__name__}: {e}. Continuing...",
                            "warning"
                        )
                    except Exception as e:
                        self.create_log.log_message(
                            self.logger,
                            f"Iteration failed due to {e.__class__.__name__}: {e}",
                            "critical"
                        )
                        raise Exception(
                            f"Iteration failed due to {e.__class__.__name__}: {e}"
                        )
                    sleep(self.int_delay_seconds)
            else:
                raise Exception(
                    "Neither {{slug}} or {{chunk_slugs}} are found in the app "
                    + "parameter, please revisit it."
                )
        else:
            while True:
                try:
                    list_ser.extend(
                        self.trt_req(
                            self.dict_metadata[str_resource]["req_method"],
                            host_,
                            self.dict_metadata[str_resource]["dtypes"],
                            dict_headers,
                            payload,
                            StrHandler().fill_placeholders(app_, {"i": i}),
                            self.dict_metadata[str_resource]["bl_verify"],
                            self.dict_metadata[str_resource]["timeout"],
                            self.dict_metadata[str_resource].get("cookies", None),
                            self.dict_metadata[str_resource]["cols_from_case"],
                            self.dict_metadata[str_resource]["cols_to_case"],
                            self.dict_metadata[str_resource]["list_cols_drop_dupl"],
                            self.dict_metadata[str_resource]["str_fmt_dt"],
                            self.dict_metadata[str_resource]["type_error_action"],
                            self.dict_metadata[str_resource]["strategy_keep_when_dupl"],
                            self.dict_metadata[str_resource].get("regex_patterns", None),
                            self.dict_metadata[str_resource].get("df_read_params", None),
                            self.dict_metadata[str_resource].get("ignored_file_extensions", []),
                            self.dict_metadata[str_resource].get("fixed_width_layout", [{}]),
                            self.dict_metadata[str_resource].get("xml_keys", None),
                            self.dict_metadata[str_resource].get("fillna_strt", {}),
                            self.dict_metadata[str_resource]["web_driver"].get(
                                "xpath_el_wait_until_loaded", {}),
                            self.dict_metadata[str_resource].get("xpaths", [{}]),
                        ).to_dict(orient="records")
                    )
                    if self.cls_db is not None:
                        self.insert_table(
                            str_resource,
                            list_ser,
                            self.dict_metadata[str_resource]["bl_insert_or_ignore"],
                            self.dict_metadata[str_resource]["bl_schema"],
                        )
                        if bl_fetch == False:
                            list_ser = list()
                    i += 1
                except tuple(list_ignorable_exceptions) as e:
                    self.create_log.log_message(
                        self.logger,
                        "Iteration encountered an ignorable exception "
                        + f"{e.__class__.__name__}: {e}. Continuing...",
                        "warning"
                    )
                except Exception as e:
                    self.create_log.log_message(
                        self.logger,
                        f"Iteration failed due to {e.__class__.__name__}: {e}",
                        "critical"
                    )
                    raise Exception(
                        f"Iteration failed due to {e.__class__.__name__}: {e}"
                    )
                sleep(self.int_delay_seconds)
        if len(list_ser) > 0:
            return pd.DataFrame(list_ser)
        else:
            return None

    def source(
        self,
        str_resource: str,
        host: Optional[str] = None,
        bl_fetch: bool = False,
    ) -> Optional[pd.DataFrame]:
        """
        Get/load raw data - if a class database is passed, the data collected will be inserted into
            the respective table; please make sure this key is filled within the metadata
        Args:
            str_resource (str): resource name
            host (Optional[str]): host name
            bl_fetch (bool): False as default
        Returns:
            Optional[pd.DataFrame]
        """
        dict_instance_vars = self.get_instance_variables
        app_ = self.dict_metadata[str_resource].get("app", None)
        if (app_ is not None) and (
            StrHandler().match_string_like(app_, "*{{*}}*") == True
        ):
            app_ = StrHandler().fill_placeholders(app_, dict_instance_vars)
        if (app_ is not None) and (
            (StrHandler().match_string_like(app_, "*{{i}}*") == True)
            or (StrHandler().match_string_like(app_, "*{{slug}}*") == True)
            or (StrHandler().match_string_like(app_, "*{{chunk_slugs}}*") == True)
        ):
            func_get_data = self.iteratively_get_data
        else:
            func_get_data = self.non_iteratively_get_data
        return func_get_data(str_resource, host, bl_fetch)



import pandas as pd
from datetime import datetime
from typing import Optional, List
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_WW_INVESTINGCOM
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests


class InvestingCom(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_start:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 5),
        dt_end:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 0),
        str_ticker:str='PETR4',
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None,
        token:Optional[str]=None,
        list_slugs:Optional[List[str]]=None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_WW_INVESTINGCOM,
            session=session,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_start = dt_start
        self.dt_end = dt_end
        self.cls_db = cls_db
        self.logger = logger
        self.token = token,
        self.list_slugs = list_slugs
        self.ticker = str_ticker
        self.dt_inf_unix_ts = DatesBR().datetime_to_unix_timestamp(dt_start)
        self.dt_sup_unix_ts = DatesBR().datetime_to_unix_timestamp(dt_end)
        self.ticker_id = self.source('ticker_id', bl_fetch=True)['ticker_id'].tolist()[0]

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        return None



### METHODS TO HANDLE STRINGS ###

from base64 import b64encode


class ImgHandler:

    def img_to_html(self, complete_path, format='jpeg'):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # convert imagem to base64
        data_uri = b64encode(open(complete_path, 'rb').read()).decode('utf-8')
        # create tag
        img_tag = '<img src="data:image/{};base64,{}">'.format(format, data_uri)
        # returning image tag
        return img_tag



# HANDLING NUMERCIAL SEQUENCES IN PYTHON

from scipy.misc import derivative
import math


class Fibonacci:

    def __init__(self, set_cache={0: 0, 1: 1}):
        self.set_cache = set_cache

    def fibonacci_of_n(self, n):
        """
        REFERENCES: https://realpython.com/fibonacci-sequence-python/
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # base case
        if n in self.set_cache:
            return self.set_cache[n]
        # compute and cache the fibonacci number
        self.set_cache[n] = self.fibonacci_of_n(
            n - 1) + self.fibonacci_of_n(n - 2)
        # recursive case
        return self.set_cache[n]

    def fibonacci(self, n):
        """
        REFERENCES: https://realpython.com/fibonacci-sequence-python/
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return [self.fibonacci_of_n(i) for i in range(n)]


class TaylorSeries:
    """
    REFERENCES: https://www.anthonymorast.com/blog/2021/10/03/taylor-series-in-python/#:~:text=The%20Taylor%20series%20of%20a,and%20using%20higher%20degree%20polynomials.
    DOCSTRING:
    INPUTS:
    OUTPUTS:
    """

    def __init__(self, function, order, center=0):
        self.center = center
        self.f = function
        self.order = order
        self.d_pts = order * 2
        self.coefficients = []
        # number of points (order) for scipy.misc.derivative
        # must be odd and greater than derivative order
        if self.d_pts % 2 == 0:
            self.d_pts += 1
        # find taylor series coefficients
        self.__find_coefficients()

    def __find_coefficients(self):
        """
        DOCSTRING: FIND TAYLOR SERIES COEFFICIENTS
        INPUTS: -
        OUTPUTS: -
        """
        for i in range(0, self.order + 1):
            self.coefficients.append(round(derivative(
                self.f, self.center, n=i, order=self.d_pts) / math.factorial(i), 5))

    def print_equation(self):
        """
        DOCSTRING: PRINT TAYLOR SERIES EQUATION
        INPUTS: -
        OUTPUTS: -
        """
        eqn_string = ''
        for i in range(self.order + 1):
            if self.coefficients[i] != 0:
                eqn_string += str(self.coefficients[i]) + (
                    '(x-{})^{}'.format(self.center, i) if i > 0 else '') + ' + '
        eqn_string = eqn_string[:-
                                3] if eqn_string.endswith(' + ') else eqn_string
        print(eqn_string)

    def print_coefficients(self):
        """
        DOCSTRING: PRINT TAYLOR SERIES' COEFFICIENTS
        INPUTS: -
        OUTPUTS: -
        """
        print(self.coefficients)

    def approximate_value(self, x):
        """
        DOCSTRING: APPROXIMATES THE VALUE OF F(X) USING THE TAYLOR POLYNOMIAL.
            X = POINT TO APPROXIMATE F(X)
        INPUTS: FLOAT
        OUTPUT: FLOAT
        """
        fx = 0
        for i in range(len(self.coefficients)):
            # coefficient * nth term
            fx += self.coefficients[i] * ((x - self.center)**i)
        return fx

    def approximate_derivative(self, x):
        """
        DOCSTRING: ESTIMATES THE DERIVATIVE OF A FUNCTION F(X) FROM ITS TAYLOR SERIES.
            USELESS SINCE WE NEED THE DERIVATIVE OF THE ACTUAL FUNCTION TO FIND THE SERIES
        INPUTS: FLOAT
        OUTPUTS: FLOAT
        """
        value = 0
        # coefficient * nth term
        for i in range(1, len(self.coefficients)):
            #   differentiate each term: x^n => n*x^(n-1)
            value += self.coefficients[i] * i * ((x - self.center)**(i - 1))
        return value

    def approximate_integral(self, x0, x1):
        """
        DOCSTRING: ESTIMATES THE DEFINITE INTEGRAL OF THE FUNCTION USING THE TAYLOR SERIES EXPANSION.
            MORE USEFUL, CONSIDER E^X * SIN(X), EASY TO DIFFERENTIATE BUT DIFFICULT TO INTEGRATE.
            X0 - LOWER LIMIT OF INTEGRATION
            X1 - UPPER LIMIT OF INTEGRATION
        INPUTS: X0, X1
        OUTPUT: FLOAT
        """

        # integrals can be off by a constant since int(f(x)) = F(x) + C
        value = 0
        for i in range(len(self.coefficients)):
            #   integrate each term: x^n => (1/n+1)*x^(n+1)
            value += ((self.coefficients[i] * (1 / (i + 1)) * ((x1 - self.center)**(i + 1))) -
                      (self.coefficients[i] * (1 / (i + 1)) * ((x0 - self.center)**(i + 1))))
        return value

    def get_coefficients(self):
        """
        DOCSTRING: RETURNS THE COEFFICIENTS OF THE TAYLOR SERIES
        INPUTS: -
        OUTPUTS: -
        """
        return self.coefficients



import logging
import os
import time
import inspect
from typing import Literal, Optional
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class CreateLog(metaclass=TypeChecker):


    def creating_parent_folder(self, new_path: str):
        if not os.path.exists(new_path):
            os.makedirs(new_path)
            return True
        else:
            return False

    def basic_conf(self, complete_path: str, basic_level: str ="info") -> logging.Logger:
        if basic_level == 'info':
            level = logging.INFO
        elif basic_level == 'debug':
            level = logging.DEBUG
        else:
            raise Exception(
                'Level was not properly defined in basic config of logging, please check')
        logging.basicConfig(
            level=level,
            filename=complete_path,
            format=(
                '%(asctime)s.%(msecs)03d %(levelname)s {%(module)s} [%(funcName)s] '
                '%(message)s'
            ),
            datefmt='%Y-%m-%d,%H:%M:%S',
        )
        console = logging.StreamHandler()
        console.setLevel
        logger = logging.getLogger(__name__)
        return logger

    def info(self, logger: Optional[logging.Logger], msg_str: str) -> logging.Logger:
        return logger.info(msg_str)

    def warning(self, logger: Optional[logging.Logger], msg_str: str) -> logging.Logger:
        return logger.warning(msg_str)

    def error(self, logger: Optional[logging.Logger], msg_str: str) -> logging.Logger:
        return logger.error(msg_str, exc_info=True)

    def critical(self, logger: Optional[logging.Logger], msg_str: str) -> logging.Logger:
        return logger.error(msg_str)

    def log_message(self, logger: Optional[logging.Logger], message: str,
                    log_level: Literal["info", "warning", "error", "critical"]) -> None:
        """
        Unified logging method that works with all CreateLog levels.

        Args:
            message: The log message
            log_level: One of 'info', 'warning', 'error', 'critical' (matches CreateLog methods)

        Returns:
            None
        """
        frame = inspect.currentframe()
        class_name = 'UnknownClass'
        method_name = 'unknown_method'
        set_skip_modules = {
            'pydantic',
            'typing',
            'inspect',
            'logging',
            'stpstone.transformations.validation'
        }
        # walk up the call stack to find the first non-CreateLog frame
        while frame:
            frame = frame.f_back
            if not frame:
                break
            module_name = frame.f_globals.get('__name__', 'UnknownModule')
            if any(module_name.startswith(prefix) for prefix in set_skip_modules):
                continue
            self_potential_cls = frame.f_locals.get('self')
            if self_potential_cls is not None and not isinstance(self_potential_cls, CreateLog):
                class_name = self_potential_cls.__class__.__name__
                method_name = frame.f_code.co_name
                break
            # fallback to function name if no suitable self found
            method_name = frame.f_code.co_name
        formatted_message = f"[{class_name}.{method_name}] {message}"
        if logger is not None:
            log_method = getattr(self, log_level, self.locals()[log_level])
            log_method(logger, formatted_message)
        else:
            level = log_level.upper()
            timestamp = f"{time.strftime('%Y-%m-%d,%H:%M:%S')}.{int(time.time() * 1000) % 1000:03d}"
            print(f"{timestamp} {level} {{{class_name}}} [{method_name}] {message}")


def timeit(method: callable) -> callable:
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result
    return timed

def conditional_timeit(bl_use_timer: bool) -> callable:
    """
    Applies the @timeit decorator conditionally based on `use_timer`

    Args:
        use_timer: boolean indicating whether to apply timing.

    Returns:
        a function wrapped with the @timeit decorator if `use_timer` is true.
    """
    def decorator(method):
        if bl_use_timer:
            return timeit(method)
        return method
    return decorator



import pandas as pd
from datetime import datetime
from typing import Optional
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from stpstone._config.global_slots import YAML_WW_CRYPTO_COINPAPRIKA
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests


class CoinPaprika(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None
    ) -> None:
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        super().__init__(
            dict_metadata=YAML_WW_CRYPTO_COINPAPRIKA,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger
        )

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        return None



import pandas as pd
from datetime import datetime
from typing import Optional, List
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_B3_BVMF_BOV
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.numbers import NumHandler


class BVMFBOV(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None,
        token:Optional[str]=None,
        list_slugs:Optional[List[str]]=None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_B3_BVMF_BOV,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token,
        self.list_slugs = list_slugs
        self.dt_ref_mm_yyyy = dt_ref.strftime('%m/%Y')

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        # setting variables
        list_th = list()
        list_td = list()
        bs_html = HtmlHandler().bs_parser(resp_req)
        # table
        bs_table = bs_html.find_all('table')[11]
        # looping within rows
        for i_tr, tr in enumerate(bs_table.find_all('tr')):
            #   checking if is a header how, otherwise it is a data row
            if i_tr < 2:
                #   getting headers
                list_th.extend([
                    StrHandler().remove_diacritics(el.get_text())
                        .replace('\xa0', '')
                        .replace('Totais dos pregoes  Ref: ', '')
                        .replace('(R$)', 'BRL')
                        .replace(' - ', ' ')
                        .replace(' ', '_')
                        .strip()
                        .upper()
                    for el in tr.find_all('td')
                    if len(
                        StrHandler().remove_diacritics(el.get_text())\
                        .replace('\xa0', '')
                    ) > 0
                ])
            else:
                #   getting data
                list_td.extend([
                    # data numeric
                    float(td.get_text()
                        .strip()
                        .replace('.', '')
                        .replace(",", "."))
                    if NumHandler().is_numeric(
                        StrHandler().remove_diacritics(td.get_text())
                            .strip()
                            .replace('.', '')
                            .replace(",", ".")
                    )
                    # data not numeric
                    else
                        StrHandler().remove_diacritics(td.get_text())
                            .strip()
                            .replace(' de ', ' ')
                            .replace(' do ', ' ')
                            .replace(' a ', ' ')
                            .replace(' e ', ' ')
                            .replace(' - ', ' ')
                            .replace('-', ' ')
                            .replace(' / ', ' ')
                            .replace(' ', '_')
                            .replace('.', '')
                            .replace(",", ".")
                            .replace('(', '')
                            .replace(')', '')
                            .replace('/', '')
                            .upper()
                    for td in tr.find_all('td')
                ])
        # dealing with header raw data
        list_th = [
            list_th[2],
            list_th[3],
            list_th[4],
            list_th[3] + '_' + '12M',
            list_th[4] + '_' + '12M',
        ]
        # pair headers and data within a list
        list_ser = HandlingDicts().pair_headers_with_data(
            list_th,
            list_td
        )
        # turning into dataframe
        df_ = pd.DataFrame(list_ser)
        df_['PERIODO_REF'] = self.dt_ref_mm_yyyy
        return df_



import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Tuple
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from bs4 import BeautifulSoup
from stpstone._config.global_slots import YAML_WW_GR
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.numbers import NumHandler
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.lists import ListHandler


class GlobalRates(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_WW_GR,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs

    def repeat_list_n_times(self, list_: List[Any], n: int) -> List[Any]:
        return list_ * n

    def repeat_elements_in_sequence(self, list_: List[Any], n: int) -> List[Any]:
        return [element for element in list_ for _ in range(n)]

    def get_td_th_raw(self, bs_table: BeautifulSoup) -> pd.DataFrame:
        list_outlooks_raw = list()
        list_th_raw = [
            th.get_text().replace("\n", "").replace("/", "_").strip().upper()
            for th in bs_table.find_all("th")
        ]
        list_td_raw = [
            float(td.get_text().replace("\n", "").replace("%", "").strip()) / 100.0
            if NumHandler().is_numeric(td.get_text().replace("\n", "").replace("%", "").strip())
            else td.get_text().replace("\n", "").replace("%", "").strip()
            for td in bs_table.find_all("td") if len(td.get_text().replace("\n", "")\
                                                     .replace("%", "").replace("-", "0").strip()) > 0
        ]
        try:
            for div in bs_table.find_all("div", class_="table-normal text-end"):
                try:
                    str_outlook = StrHandler().replace_all(
                        div.find("i")["class"][1], {
                            "fa-circle-down": "DOWNWARD",
                            "fa-circle-up": "UPWARD"
                        }
                    )
                    list_outlooks_raw.append(str_outlook)
                except (AttributeError, TypeError):
                    continue
            int_col_dir = list_th_raw.index("DIRECTION")
            int_rng_upper = len(list_td_raw) // (len(list_th_raw) - 1)
            for i in range(int_rng_upper):
                list_td_raw.insert(3 + (int_col_dir + 3) * i, list_outlooks_raw[i])
        except ValueError:
            pass
        return list_th_raw, list_td_raw, list_outlooks_raw

    def trt_td_th_raw(self, list_th_raw: List[Any], list_td_raw: List[Any], str_source: str) \
        -> pd.DataFrame:
        list_th_dts = [x for x in list_th_raw if len(x) > 0]
        list_td_rts_names = [x for x in list_td_raw if isinstance(x, str)]
        list_td_rts_values = [x for x in list_td_raw if NumHandler().is_number(x) == True]
        list_rts_names = self.repeat_elements_in_sequence(list_td_rts_names, len(list_th_dts))
        list_rt_dts = self.repeat_list_n_times(list_th_dts, len(list_td_rts_names))
        if str_source in \
            ["ester", "sonia", "sofr", "usa_cpi", "british_cpi", "canadian_cpi", "european_cpi"]:
            list_rt_dts = list_td_rts_names.copy()
            list_rts_names = [f"{str_source}".upper()] * len(list_td_rts_values)
        list_ser = {
            "DATE": list_rt_dts,
            "RATE_NAME": list_rts_names,
            "RATE_VALUE": list_td_rts_values
        }
        if str_source in ["central_banks"]:
            list_cols = ["CENTRAL_BANK"] + ListHandler().remove_duplicates(list_rt_dts)
            list_data = ListHandler().remove_consecutive_duplicates(list_rts_names)
            for i_rt_val, i in enumerate(range(2, len(list_data) + len(list_td_rts_values) + 1, 6)):
                i_rt_val_ = i_rt_val * 2
                list_data.insert(i, list_td_rts_values[i_rt_val_])
                list_data.insert(i + 2, list_td_rts_values[i_rt_val_ + 1])
            list_ser = HandlingDicts().pair_headers_with_data(list_cols, list_data)
        return list_ser

    def td_th_parser(self, bs_table: BeautifulSoup, str_source: str) -> pd.DataFrame:
        list_th_raw, list_td_raw, list_outlooks_raw = self.get_td_th_raw(bs_table)
        list_ser = self.trt_td_th_raw(list_th_raw, list_td_raw, str_source)
        return pd.DataFrame(list_ser)

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        bs_html = HtmlHandler().bs_parser(resp_req)
        str_source = StrHandler().get_url_query(resp_req.url, bl_include_fragment=True)["source"]
        bs_table = bs_html.find("table")
        return self.td_th_parser(bs_table, str_source)
